# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import ast
import multiprocessing
import json
import datasets
import socket
import queue
import dataclasses
import base64
import struct
import tarfile
import ctypes
import asyncio
import importlib
import sys
import glob
import io
from PIL import Image
from filelock import FileLock, Timeout
from collections import OrderedDict
from functools import wraps
from msgpack import ExtType
from tqdm import tqdm
from typing import List, Tuple, Dict, Any, Optional
import torch
import pynvml
from contextlib import contextmanager
from torch.functional import F
from huggingface_hub import (
    hf_hub_download,
    list_repo_files,
    snapshot_download,
    HfFileSystem,
)
import time
import functools
from cosmos_rl.utils.logging import logger
from safetensors import safe_open
from cosmos_rl.utils.constant import CACHE_DIR
import math


def create_cached_dir_if_needed():
    """
    Creates the local cached dir if it doesn't exist.
    """
    if not CACHE_DIR.exists():
        CACHE_DIR.mkdir(parents=True, exist_ok=True)


def load_data_from_disk_or_hf(data_name, data_subset=None, revision=None):
    if data_subset is not None and len(data_subset) == 0:
        data_subset = None

    if os.path.exists(data_name):
        try:
            return datasets.load_from_disk(data_name)
        except Exception as e:
            logger.warning(
                f"Failed to load dataset from disk: {e}. Trying to load from HuggingFace Hub..."
            )
            return datasets.load_dataset(data_name, data_subset, revision=revision)
    return datasets.load_dataset(data_name, data_subset, revision=revision)


def read_json_file(file_path):
    with open(file_path, "r") as file:
        json_data = json.load(file)
    return json_data


def write_json_file(data, file_path):
    with open(file_path, "w") as json_file:
        json.dump(data, json_file, indent=4)


def resolve_model_path(model_path: str, revision: Optional[str] = None) -> str:
    if not os.path.exists(model_path.replace(":", "/")):
        if ":" in model_path:
            if model_path.count(":") > 1:
                raise ValueError(
                    f"Invalid model path {model_path}, should be in the format of 'owner/repo[:path]'"
                )
            model_path, file_path = model_path.split(":")
        else:
            model_path = model_path
            file_path = None

        # Check whether `model_path` is a HuggingFace repo id with repo name
        if len(model_path.split("/")) == 2:
            logger.info(
                f"model path {model_path} is not a directory. Trying to load from HuggingFace Hub..."
            )

            hf_fs = HfFileSystem(token=os.environ.get("HF_TOKEN", None))
            files = hf_fs.ls(model_path, detail=False)
            if (
                os.path.join(model_path, "model.safetensors.index.json") in files
                or os.path.join(model_path, "model.safetensors") in files
            ):
                logger.info(
                    f"Found safetensors in {model_path}. Ignoring *pytorch_model* and *consolidated* files."
                )
                ignore_patterns = ["*pytorch_model*", "*consolidated*"]
            else:
                ignore_patterns = None

            try:
                model_path = retry(snapshot_download)(
                    model_path,
                    revision=revision,
                    token=os.environ.get("HF_TOKEN"),
                    cache_dir=os.environ.get(
                        "HF_HOME",
                        os.path.expanduser("~/.cache/huggingface/transformers/"),
                    ),
                    ignore_patterns=ignore_patterns,
                    allow_patterns=file_path,
                )
            except Exception as e:
                logger.error(f"Error: {e}")
                raise

            if file_path is not None:
                model_path = os.path.join(model_path, file_path)
            logger.info(f"Downloaded model from HuggingFace to {model_path}")
        elif os.path.isdir(model_path):
            # In case model_path is a local directory, we need to check if the file_path exists
            if file_path is not None:
                model_path = os.path.join(model_path, file_path)
        else:
            raise ValueError(
                f"Model path {model_path} is not a directory and not a valid HuggingFace repo id with repo name."
            )
    else:
        model_path = model_path.replace(":", "/")
    return model_path


def is_port_free(port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.connect_ex(("127.0.0.1", port)) != 0


def find_available_port(start_port):
    max_port = 65535  # Maximum port number
    for port in range(start_port, max_port + 1):
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.bind(("localhost", port))
                return port
        except OSError:
            continue

    raise RuntimeError("No available ports found in the specified range.")


def put_with_overwrite(q: queue.Queue, item):
    assert isinstance(q, queue.Queue), "q must be a queue.Queue, but not asyncio.Queue"
    if q.full():
        try:
            q.get_nowait()  # Remove the oldest item
        except queue.Empty:
            pass  # In case the queue becomes empty in a race condition
    q.put(item)


def extract_fields(dc_instance):
    """
    Recursively extract dataclass fields into a nested dictionary.
    Leaf nodes contain a dict with 'value', 'metadata', 'type', and 'input_type'.
    """

    def recursive_extract(instance):
        fields = {}
        for f in dataclasses.fields(instance):
            if f.metadata.get("skip_ui", False):
                continue
            value = getattr(instance, f.name)

            # Special handling for train_policy field
            if f.name == "train_policy" and hasattr(instance, "train_policy"):
                fields[f.name] = recursive_extract(
                    value
                )  # Just extract current policy's fields
                continue

            if dataclasses.is_dataclass(value):
                fields[f.name] = recursive_extract(value)
            else:
                field_data = {
                    "value": value,
                    "metadata": f.metadata,
                    "type": f.type,
                    "input_type": "checkbox" if f.type == bool else "text",
                }
                fields[f.name] = field_data
        return fields

    return recursive_extract(dc_instance)


def parse_collection(s):
    """
    Attempts to parse a string into a Python literal.

    If the string represents a list or tuple, it returns the corresponding
    Python object. Otherwise, it returns None.
    """
    try:
        result = ast.literal_eval(s)
    except (ValueError, SyntaxError):
        # The string is not a valid Python literal.
        return None

    # Check if the result is a list or tuple.
    if isinstance(result, (list, tuple)):
        return result
    else:
        return None


def list_to_b64(lst) -> str:
    # for int64_t listy to base64
    byte_data = struct.pack(f"{len(lst)}q", *lst)
    return base64.b64encode(byte_data).decode("utf-8")


def b64_to_list(b64_str) -> List[int]:
    # for base64 to int64_t list
    byte_data = base64.b64decode(b64_str)
    n = len(byte_data) // 8
    return list(struct.unpack(f"{n}q", byte_data))


def str2torch_dtype(dtype_str: str) -> torch.dtype:
    """
    Convert a string representation of a dtype to a torch.dtype.
    """
    dtype_str = dtype_str.lower()
    if dtype_str == "bfloat16":
        return torch.bfloat16
    elif dtype_str == "float16":
        return torch.float16
    elif dtype_str == "float32":
        return torch.float32
    else:
        raise ValueError(f"Unsupported dtype: {dtype_str}")


@contextmanager
def cosmos_default_dtype(dtype: torch.dtype):
    old = torch.get_default_dtype()
    torch.set_default_dtype(dtype)
    try:
        yield
    finally:
        torch.set_default_dtype(old)


class IdentityLayer(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x


@torch.compile
def _logsumexp_fp32_backward(grad_output, input, logsumexp_values):
    # grad_output shape: (...), need to unsqueeze to match input shape for broadcasting
    grad_output_expanded = grad_output.unsqueeze(-1)
    # Compute softmax: exp(input - logsumexp)
    softmax = (input - logsumexp_values.unsqueeze(-1)).exp()
    return (grad_output_expanded * softmax).to(input.dtype)


class LogSumExpFp32Backward(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input):
        logsumexp_values = torch.zeros(
            input.shape[:-1], dtype=torch.float32, device=input.device
        )
        torch.logsumexp(input, dim=-1, out=logsumexp_values)
        ctx.save_for_backward(input, logsumexp_values)
        ctx.input_dtype = input.dtype
        return (
            logsumexp_values.to(torch.bfloat16)
            if input.dtype == torch.bfloat16
            else logsumexp_values
        )

    @staticmethod
    def backward(ctx, grad_output):
        input, logsumexp_values = ctx.saved_tensors
        return _logsumexp_fp32_backward(grad_output, input, logsumexp_values)


def _logsumexp_fp32(input):
    """
    Computes the logsumexp of the input tensor in float32 precision (even if input is Bf16).
    """
    return LogSumExpFp32Backward.apply(input)


_SELECTIVE_LOG_SOFTMAX_OPTIM = False


def selective_log_softmax(logits, index):
    """
    A memory-efficient implementation of the common `log_softmax -> gather` operation.

    This function is equivalent to the following naive implementation:
    ```python
    logps = torch.gather(logits.log_softmax(-1), dim=-1, index=index.unsqueeze(-1)).squeeze(-1)
    ```

    Args:
        logits (`torch.Tensor`):
            Logits tensor of shape `(..., num_classes)`.
        index (`torch.Tensor`):
            Index tensor of shape `(...)`, specifying the positions to gather from the log-softmax output.

    Returns:
        `torch.Tensor`:
            Gathered log probabilities with the same shape as `index`.
    """
    if _SELECTIVE_LOG_SOFTMAX_OPTIM:
        # NOTE: aazzolini optimized.
        selected_logits = (
            torch.gather(logits, dim=-1, index=index.unsqueeze(-1))
            .squeeze(-1)
            .to(torch.float32)
        )
        logsumexp_values = _logsumexp_fp32(logits)
        return (selected_logits - logsumexp_values).to(
            logits.dtype
        )  # log_softmax(x_i) = x_i - logsumexp(x)

    if logits.dtype in [torch.float32, torch.float64]:
        selected_logits = torch.gather(
            logits, dim=-1, index=index.unsqueeze(-1)
        ).squeeze(-1)
        # loop to reduce peak mem consumption
        logsumexp_values = torch.stack([torch.logsumexp(lg, dim=-1) for lg in logits])
        per_token_logps = (
            selected_logits - logsumexp_values
        )  # log_softmax(x_i) = x_i - logsumexp(x)
    else:
        # logsumexp approach is unstable with bfloat16, fall back to slightly less efficent approach
        per_token_logps = []
        for row_logits, row_labels in zip(
            logits, index
        ):  # loop to reduce peak mem consumption
            row_logps = F.log_softmax(row_logits, dim=-1)
            row_per_token_logps = row_logps.gather(
                dim=-1, index=row_labels.unsqueeze(-1)
            ).squeeze(-1)
            per_token_logps.append(row_per_token_logps)
        per_token_logps = torch.stack(per_token_logps)
    return per_token_logps


def clear_weight_name(name: str) -> str:
    return name.replace("._orig_mod", "").replace("._checkpoint_wrapped_module", "")


def _extract_tgz_file(tgz_file, output_dir):
    with tarfile.open(tgz_file, "r:gz") as tar:
        tar.extractall(path=output_dir)


def basename_from_modelpath(path: str) -> str:
    path = path.strip().strip("/")
    if len(path.split("/")) > 2:
        path = "/".join(path.split("/")[-2:])
    return path


def if_use_modelscope(path: str) -> bool:
    modelscope_cache_dir = os.environ.get(
        "MODELSCOPE_CACHE",
        os.path.join(os.path.expanduser("~"), ".cache/modelscope_cache/dataset"),
    )
    return path.startswith(modelscope_cache_dir)


def prepare_cosmos_data(dataset, *args, **kwargs):
    cache_dir = os.environ.get(
        "COSMOS_CACHE", os.path.join(os.path.expanduser("~"), ".cache/cosmos/")
    )
    dataset_name = basename_from_modelpath(dataset.name)
    use_modelscope = if_use_modelscope(dataset.name)
    dataset_dir = os.path.join(
        cache_dir,
        "datasets",
        dataset_name,
        dataset.subset,
    )
    video_clips_dir = os.path.join(dataset_dir, "video_clips")

    prepared_flag = os.path.join(video_clips_dir, ".prepared")
    if os.path.exists(prepared_flag):
        logger.info(f"Dataset {dataset.name} already prepared.")
        return

    # ensure base dirs exist
    os.makedirs(video_clips_dir, exist_ok=True)

    # file-lock to prevent races
    lock_path = os.path.join(dataset_dir, "prepare.lock")
    lock = FileLock(lock_path, timeout=1800)  # wait up to 30 minutes

    try:
        with lock:
            # re-check sentinel inside the lock
            if os.path.exists(prepared_flag):
                logger.info(
                    f"Dataset {dataset.name} already prepared by another worker."
                )
                return

            # --- find .tar.gz clips ---
            re_pattern = re.compile(rf"^{re.escape(dataset.subset)}/clips/.*\.tar\.gz$")
            file_pattern = f"{dataset.subset}/clips/*.tar.gz"
            if use_modelscope:
                assert os.path.exists(dataset.name), f"{dataset.name} not found locally"
                remote_files = [
                    f.replace(dataset.name + "/", "")
                    for f in glob.glob(f"{dataset.name}/**/*.tar.gz", recursive=True)
                ]
            else:
                remote_files = list_repo_files(
                    repo_id=dataset_name,
                    repo_type="dataset",
                    revision=dataset.revision or None,
                )

            tgz_files = [f for f in remote_files if re_pattern.match(f)]

            # --- extract clips ---
            if tgz_files:
                if use_modelscope:
                    downloaded_clips_dir = os.path.join(
                        dataset.name, dataset.subset, "clips"
                    )
                else:
                    snapshot_dir = retry(snapshot_download)(
                        dataset_name,
                        allow_patterns=[file_pattern],
                        repo_type="dataset",
                        revision=dataset.revision or None,
                    )
                    downloaded_clips_dir = os.path.join(
                        snapshot_dir, dataset.subset, "clips"
                    )

                assert os.path.exists(
                    downloaded_clips_dir
                ), f"Cannot find clips at {downloaded_clips_dir}"

                # parallel extract
                results = {}
                with tqdm(total=len(tgz_files), desc="Extracting clips") as pbar:
                    with multiprocessing.Pool(
                        processes=min(multiprocessing.cpu_count(), 8)
                    ) as pool:
                        for tgz in tgz_files:
                            full_path = os.path.join(
                                downloaded_clips_dir, os.path.basename(tgz)
                            )
                            results[tgz] = pool.apply_async(
                                _extract_tgz_file,
                                (full_path, video_clips_dir),
                                callback=lambda _: pbar.update(1),
                            )
                        pool.close()
                        pool.join()

                # check for extract errors
                for tgz, res in results.items():
                    if not res.successful():
                        raise RuntimeError(f"Failed to extract {tgz}: {res.get()}")

            else:
                # legacy single-archive format
                clip_tgz = os.path.join(dataset.subset, "clips.tar.gz")
                if use_modelscope:
                    clip_tgz = os.path.join(dataset.name, clip_tgz)
                else:
                    clip_tgz = hf_hub_download(
                        repo_id=dataset_name,
                        revision=dataset.revision or None,
                        repo_type="dataset",
                        filename=clip_tgz,
                    )
                _extract_tgz_file(clip_tgz, video_clips_dir)

            # --- special renaming for 'av' subset ---
            if dataset.subset == "av":
                for root, _, files in os.walk(video_clips_dir):
                    for file in files:
                        if file.endswith(".mp4"):
                            src = os.path.join(root, file)
                            dst = os.path.join(root, file.split(".")[0] + ".mp4")
                            os.rename(src, dst)
                            logger.info(f"Renamed {file} → {os.path.basename(dst)}")

            # finally, create our sentinel
            open(prepared_flag, "w").close()
            logger.info(f"Finished preparing {dataset.name}.")

    except Timeout:
        raise RuntimeError(
            f"Timeout acquiring lock for {dataset.name}; another process may be stuck."
        )


GPU_FLOPS_MAPPING = {
    "NVIDIA H100 80GB HBM3": {
        "FP32": 989 * (1e12),
        "FP16": 1979 * (1e12),
    },
    "NVIDIA A100 80GB": {
        "FP32": 39 * (1e12),
        "FP16": 195 * (1e12),
    },
}


def get_device_flops(dtype: torch.dtype, num_gpus: int) -> int:
    """
    Get the GPU FLOPs for the current device.

    Args:
        dtype (torch.dtype): The data type of the model.
        num_gpus (int): The number of GPUs available.

    Returns:
        int: The FLOPs of the current device.
    """
    gpu_flops = 0
    if torch.cuda.is_available():
        pynvml.nvmlInit()
        handle = pynvml.nvmlDeviceGetHandleByIndex(0)
        gpu_type = pynvml.nvmlDeviceGetName(handle).decode("utf-8")
        if dtype == "float32":
            gpu_flops = GPU_FLOPS_MAPPING[gpu_type]["FP32"]
        elif dtype == "float16" or dtype == "bfloat16":
            gpu_flops = GPU_FLOPS_MAPPING[gpu_type]["FP16"]
        else:
            raise ValueError(f"Unsupported dtype: {dtype}")
    else:
        logger.warning("CUDA is not available. Cannot get GPU FLOPs.")
    return gpu_flops * num_gpus


def compute_mfu(
    model: torch.nn.Module,
    n_tokens: int,
    iter_time: float,
    num_gpus: int,
    dtype: str,
) -> dict:
    """
    Compute the model FLOPs Utilization (MFU) for a given model and number of tokens.

    Args:
        model (torch.nn.Module): The model for which to compute the MFU.
        n_tokens (int): The number of tokens in the inputs.
        iter_time (float): The time taken for the forward pass in seconds.

    Returns:
        dict: A dictionary containing the MFU/FLOPs value.
    """
    result = {}
    # Get the model FLOPs for the iteration
    model_params, model_flops = model.get_nparams_and_flops(n_tokens)
    result["model_flops"] = model_flops

    # Get the GPU FLOPs
    try:
        gpu_flops = get_device_flops(dtype, num_gpus)
        # Calculate and return the MFU
        mfu = (model_flops / gpu_flops) / iter_time
        result["mfu"] = mfu
        logger.info(
            f"MFU: {mfu:.4f}, Model FLOPs: {model_flops / 1e12:.2f}T, GPU FLOPs: {gpu_flops / 1e12:.2f}T, Iter time: {iter_time:.2f}s"
        )
    except Exception as e:
        logger.error(
            f"Cannot compute MFU: {e}, only report model FLOPs, you can calculate MFU manually."
        )
        logger.info(
            f"Model FLOPs: {model_flops / 1e12:.2f}T, Iter time: {iter_time:.2f}s"
        )

    return result


def fix_data_type_size(obj):
    if isinstance(obj, tuple):
        return tuple([fix_data_type_size(x) for x in obj])
    elif isinstance(obj, list):
        return [fix_data_type_size(x) for x in obj]
    elif isinstance(obj, dict):
        return {fix_data_type_size(k): fix_data_type_size(v) for k, v in obj.items()}
    elif isinstance(obj, int):
        return ctypes.c_int64(obj)
    else:
        return obj


# Extension code
MSGPACK_C_LONG_EXT_TYPE = 0x10


# Hooks
def msgpack_c_long(obj):
    if isinstance(obj, ctypes._SimpleCData) and isinstance(obj.value, int):
        if isinstance(obj, ctypes.c_long):
            return ExtType(
                MSGPACK_C_LONG_EXT_TYPE, obj.value.to_bytes(8, "big", signed=True)
            )
    raise TypeError(f"Unsupported type: {type(obj)}")


def msgunpack_c_long(code, data):
    if code == MSGPACK_C_LONG_EXT_TYPE:
        val = int.from_bytes(data, "big", signed=True)
        return int(val)
    return ExtType(code, data)


def sync_model_vocab(
    model_name_or_path,
    lm_head_key="lm_head.weight",
    embed_tokens_key="model.embed_tokens.weight",
):
    self_rank = int(os.environ.get("RANK", 0))
    vocab_size = None
    if self_rank == 0:
        model_index_path = f"{model_name_or_path}:model.safetensors.index.json"
        weight_map_path = resolve_model_path(model_index_path)
        has_weight_map = os.path.exists(weight_map_path)
        # Check if the weight map file exists
        if has_weight_map:
            weight_map = read_json_file(weight_map_path)["weight_map"]
            possible_prefix = ["", "model.", "language_model."]
            possible_keys = [prefix + lm_head_key for prefix in possible_prefix] + [
                prefix + embed_tokens_key for prefix in possible_prefix
            ]
            for key in possible_keys:
                if key in weight_map:
                    head_or_embed_tokens = weight_map[key]
                    head_or_embed_tokens_path = resolve_model_path(
                        f"{model_name_or_path}:{head_or_embed_tokens}"
                    )
                    with safe_open(
                        head_or_embed_tokens_path, framework="pt", device="cpu"
                    ) as f:
                        tensor_slice = f.get_slice(key)
                    vocab_size, _ = tensor_slice.get_shape()
                    break
        else:
            # models like google/gemma-3-1b-pt does not have model.safetensors.index.json
            model_safetensors_path = resolve_model_path(
                f"{model_name_or_path}:model.safetensors"
            )
            with safe_open(model_safetensors_path, framework="pt", device="cpu") as f:
                tensor_names = f.keys()
                possible_prefix = ["", "model.", "language_model."]
                possible_keys = [prefix + lm_head_key for prefix in possible_prefix] + [
                    prefix + embed_tokens_key for prefix in possible_prefix
                ]
                for key in possible_keys:
                    if key in tensor_names:
                        tensor_slice = f.get_slice(key)
                        vocab_size, _ = tensor_slice.get_shape()
                        break
        if vocab_size is None:
            raise ValueError(
                "Could not find `lm_head` or `model.embed_tokens.weight` in the model."
            )

    from cosmos_rl.utils.distributed import broadcast_object_cpu

    vocab_size = broadcast_object_cpu(vocab_size, src=0, device=torch.device("cpu"))
    logger.info(f"Vocabulary size: {vocab_size}")

    return vocab_size


def retry(func=None, *, max_retry=10, max_delay=30.0):
    """
    Decorator (or wrapper) to retry a function up to max_retry times,
    backing off exponentially (1s, 2s, 4s, …) up to max_delay seconds.

    Usage:

      @retry(max_retry=5)                # uses default max_delay=30
      def foo(...): …

      @retry(max_retry=5, max_delay=60)  # override max_delay
      def bar(...): …

      wrapped = retry(baz, max_retry=2)  # direct call style
    """

    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            delay = 1.0
            for attempt in range(max_retry + 1):
                try:
                    return f(*args, **kwargs)
                except Exception:
                    if attempt == max_retry:
                        # out of retries: re-raise last exception
                        raise
                    time.sleep(delay)
                    delay = min(delay * 2, max_delay)

        return wrapper

    # allow both @retry(...) and retry(func, ...)
    if callable(func):
        return decorator(func)
    return decorator


def write_redis_config(port, logfile, file_path="/opt/redis_config.conf"):
    """
    Write the redis config file.
    redis_config_path: the path to the redis config file.
    port: the port for Redis to listen on.
    logfile: the logfile for Redis.

    return the actual path of the redis config file.
    """
    config_content = f"""# Redis configuration file example for insecure connections

# Bind to all network interfaces (use with caution)
bind 0.0.0.0

# Set the port for Redis to listen on (default is {port})
port {port}

# Disable TLS by setting the tls-port to 0
tls-port 0

# Disable authentication by commenting out the requirepass directive
# requirepass yourpassword

# Other configuration settings can remain as default or be customized as needed
timeout 0
tcp-keepalive 300
protected-mode no
# enable-protected-configs yes
# enable-debug-command yes
# enable-module-command yes
daemonize yes
supervised no
loglevel notice
logfile {logfile}
databases 16
save 900 1
save 300 10
save 60 10000
stop-writes-on-bgsave-error yes
rdbcompression yes
rdbchecksum yes
dbfilename dump.rdb
dir /opt
"""
    with open(file_path, "w") as file:
        file.write(config_content)
    return file_path


def do_once(func):
    """Decorator to make a function only call once."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        key = func.__name__ + str(args) + str(kwargs)
        if key not in do_once.dones:
            do_once.dones.add(key)
            return func(*args, **kwargs)

    return wrapper


do_once.dones = set()


def is_master_rank(parallel_dims, global_rank: int):
    return (not parallel_dims.pp_enabled and global_rank == 0) or (
        parallel_dims.pp_enabled
        and global_rank
        == (parallel_dims.world_size - parallel_dims.world_size / parallel_dims.pp)
    )


def masked_mean(values, mask, axis=None):
    """
    Compute the mean of `values` over elements selected by `mask`.

    Args:
        values (Tensor): Input tensor.
        mask (Tensor): Boolean or numeric mask of the same shape as `values`.
        axis (int or tuple of int, optional): Dimension(s) along which to compute the mean.
            Defaults to None (over all elements).

    Returns:
        Tensor: Masked mean, with shape equal to `values` reduced over `axis`.
    """
    return (values * mask).sum(axis=axis) / (mask.sum(axis=axis) + 1e-8)


def is_cuda_compatible(major: int, minor: int) -> bool:
    return torch.cuda.is_available() and torch.cuda.get_device_capability() >= (
        major,
        minor,
    )


# From torchao.
def parse_version(version_string):
    # Extract just the X.Y.Z part from the version string
    match = re.match(r"(\d+\.\d+\.\d+)", version_string)
    if match:
        version = match.group(1)
        return [int(x) for x in version.split(".")]
    else:
        raise ValueError(f"Invalid version string format: {version_string}")


def compare_versions(v1, v2):
    v1_parts = parse_version(v1)
    v2_parts = parse_version(v2)
    return (v1_parts > v2_parts) - (v1_parts < v2_parts)


def is_fbcode():
    return not hasattr(torch.version, "git_version")


def torch_version_at_least(min_version):
    return is_fbcode() or compare_versions(torch.__version__, min_version) >= 0


class TrieNode:
    __slots__ = ("children", "idxs")

    def __init__(self):
        self.children = {}
        self.idxs = []


def build_trie(seqs):
    root = TrieNode()
    for i, s in enumerate(seqs):
        node = root
        for x in s:
            node.idxs.append(i)
            node = node.children.setdefault(x, TrieNode())
        node.idxs.append(i)
    return root


def find_maximal_prefix_groups(
    seqs: List[List[int]],
    N: int,
):
    if N is None or N < 1:
        return {}
    root = build_trie(seqs)
    result = {}
    # We'll do a post-order traversal with an explicit stack
    # Each frame is (node, prefix, visited_flag)
    stack = [(root, (), False)]
    while stack:
        node, prefix, visited = stack.pop()
        if not visited:
            # push back as “to be processed after children”
            stack.append((node, prefix, True))
            # push children
            for x, child in node.children.items():
                stack.append((child, prefix + (x,), False))
        else:
            # now all children have been “visited” so we can decide if
            # this node is a maximal group
            if len(prefix) >= N and len(node.idxs) > 1:
                # check if any child also has ≥2 idxs
                has_deeper = any(
                    len(child.idxs) > 1 for child in node.children.values()
                )
                if not has_deeper:
                    result[prefix] = list(node.idxs)
    return result


def add_nan_checks(model):
    """
    Add nan checks to the model.
    """
    for name, param in model.named_parameters():
        if not param.requires_grad or not param.is_leaf:
            continue

        # factory to capture the current name in a closure
        def make_hook(param_name):
            def hook(grad):
                origin_grad = grad
                if isinstance(grad, torch.distributed.tensor.DTensor):
                    grad = grad.to_local()
                if torch.isnan(grad).any():
                    msg = f"NaN detected in gradient of {param_name}"
                    raise RuntimeError(msg)
                return origin_grad  # must return the (possibly modified) grad

            return hook

        param.register_hook(make_hook(name))
        logger.info(f"Added nan check for {name}")


# Util func to create an asyncio task with proper cleanup
strong_refs = set()


def create_async_task(coro):
    task = asyncio.create_task(coro)
    strong_refs.add(task)
    task.add_done_callback(strong_refs.discard)
    return task


def compute_logprobs(
    input_ids_batch: torch.Tensor,  # [batch_size, max_len]
    logprob_masks: torch.Tensor,  # [batch_size, max_len],
    logits: torch.Tensor,  # [batch_size, max_len, vocab_size] or [n_logprob_tokens, vocab_size] if is_full_logits is False
    is_full_logits: bool = False,
    label_packing_mask: Optional[torch.Tensor] = None,  # [batch_size, max_len]
    input_packing_mask: Optional[torch.Tensor] = None,  # [batch_size, max_len]
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Compute the per-token log probabilities and advantages

    Args:
        input_ids_batch: the input_ids of the model [batch_size, max_len]
        logprob_masks: the logprob_masks of the model [batch_size, max_len]
        logits: the logits of the model [batch_size, max_len, vocab_size] or [n_logprob_tokens, vocab_size]
        is_full_logits: whether the logits are full logits or have been index-selected for memory efficiency

    Returns:
        logps: the per-token log probabilities
        cu_seqlens: the cumulative sequence lengths of the logps
    """
    # Shift token_ids
    if label_packing_mask is not None:
        assert (
            input_packing_mask is not None
        ), "input_packing_mask must be provided if label_packing_mask is used"
        shifted_input_ids = torch.zeros_like(input_ids_batch)
        shifted_input_ids[input_packing_mask] = input_ids_batch[label_packing_mask]
    else:
        shifted_input_ids = torch.empty_like(input_ids_batch)
        shifted_input_ids[:, :-1] = input_ids_batch[:, 1:]
        shifted_input_ids[:, -1] = 0

    if is_full_logits:
        assert (
            logits.shape[:2] == shifted_input_ids.shape[:2]
        ), f"Logits shape {logits.shape} does not match input_ids shape {shifted_input_ids.shape}"
        effective_logits = logits[logprob_masks]
    else:
        effective_logits = logits
    bsz, _ = input_ids_batch.shape

    effective_input_ids = shifted_input_ids[logprob_masks]  # [n_logprob_tokens,]
    # Compute sft cross entropy loss
    # sft_loss = torch.nn.functional.cross_entropy(
    #     effective_logits.view(-1, effective_logits.size(-1)),
    #     effective_input_ids.view(-1),
    #     reduction="mean",
    # )
    # print(f"[Policy] SFT loss: {sft_loss}, shape: {effective_input_ids.shape, effective_logits.shape}")

    masked_seqlens = logprob_masks.sum(dim=-1)  # [bsz,]
    cu_seqlens = torch.zeros(
        bsz + 1, dtype=torch.int32, device=logits.device
    )  # [bsz + 1,]
    cu_seqlens[1:] = torch.cumsum(masked_seqlens, dim=0)
    logps = selective_log_softmax(
        effective_logits, effective_input_ids
    )  # [n_logprob_tokens,]
    return logps, cu_seqlens


def dynamic_import_module(path: str, attr: Optional[str] = None) -> Dict[str, Any]:
    """
    Dynamically import either:
        - a single .py file
        - a package directory (must contain __init__.py)
    and allow it to use relative imports internally.

    Args:
        path: the path to the module to import
        attr: the attribute to import from the module, can be recursive attribute like `model.attr_a.attr_b.attr_c....`

    Returns the imported module object.
    """
    if os.path.exists(path):
        path = os.path.abspath(path)
        if os.path.isdir(path):
            # it's a package dir
            pkg_dir = path
            if not os.path.isfile(os.path.join(pkg_dir, "__init__.py")):
                raise ImportError(f"{pkg_dir!r} is not a package (no __init__.py)")
            module_name = os.path.basename(pkg_dir)
            parent_dir = os.path.dirname(pkg_dir)
        else:
            # it's a single .py file
            if not path.lower().endswith(".py"):
                raise ImportError(
                    f"{path!r} is neither a .py file nor a package directory"
                )
            parent_dir, filename = os.path.split(path)
            module_name = os.path.splitext(filename)[0]
        # Ensure the parent directory is on sys.path
        if parent_dir not in sys.path:
            sys.path.insert(0, parent_dir)
    else:
        # Direct python module
        module_name = path

    # Now import by name – normal import machinery applies
    module = importlib.import_module(module_name)

    if attr:
        obj = module
        for attr_part in attr.split("."):
            if not hasattr(obj, attr_part):
                raise ImportError(f"Attribute {attr} not found in {path}")
            obj = getattr(obj, attr_part)
        return obj
    return module


class RollingDict(OrderedDict):
    def __init__(self, maxlen=20):
        super().__init__()
        self.maxlen = maxlen

    def __setitem__(self, key, value):
        if key in self:
            del self[key]
        super().__setitem__(key, value)
        if len(self) > self.maxlen:
            self.popitem(last=False)


def sanitize(obj):
    if isinstance(obj, float):
        return obj if math.isfinite(obj) else None
    elif isinstance(obj, dict):
        return {k: sanitize(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [sanitize(v) for v in obj]
    else:
        return obj


def safe_deep_getattr(obj, attr_path, default=None):
    try:
        for attr in attr_path.split("."):
            obj = getattr(obj, attr)
        return obj
    except AttributeError:
        return default


def load_model_class_by_config(hf_config):
    architectures = hf_config.architectures
    assert (
        len(architectures) == 1
    ), f"zero or multiple architectures in config.json is not supported, {architectures=}"

    class_name = architectures[0]
    model_class = None
    try:
        model_class = getattr(importlib.import_module("transformers"), class_name)
    except Exception as e:
        logger.error(f"Can not load {class_name} from transformers")
        raise e
    return model_class


def reverse_hf_checkpoint_mapping(mapping):
    reverse_map = {}
    for k, v in mapping.items():
        reverse_map[r"^" + re.escape(v)] = k.lstrip("^")
    return reverse_map


def decode_vision_info(prompts):
    new_prompts = []
    for prompt in prompts:
        new_prompt = prompt.copy()
        multi_modal_data = new_prompt.pop("multi_modal_data")
        if "image" in multi_modal_data:
            img_obj = multi_modal_data["image"]
            assert isinstance(
                img_obj, list
            ), f"image should be a list, but got {type(img_obj)}"
            img_type = type(img_obj[0])
            if img_type == str:
                for i, img_b64 in enumerate(img_obj):
                    img_bytes = base64.b64decode(img_b64)
                    img_buffer = io.BytesIO(img_bytes)
                    image = Image.open(img_buffer)
                    multi_modal_data["image"][i] = image
            elif img_type == torch.Tensor:
                pass
            else:
                raise ValueError(f"Unsupported image type: {img_type}")
        if "video" in multi_modal_data:
            video_obj = multi_modal_data["video"]
            assert isinstance(
                video_obj, list
            ), f"video should be a list, but got {type(video_obj)}"
            video_type = type(video_obj[0])
            # No need to decode video tensor
            if video_type == torch.Tensor:
                pass
            else:
                raise ValueError(f"Unsupported video type: {video_type}")

        new_prompt["multi_modal_data"] = multi_modal_data
        new_prompts.append(new_prompt)
    return new_prompts


def rank0_print(msg, *args, **kwargs):
    if os.environ.get("RANK", "0") == "0":
        logger.info(msg, *args, **kwargs)


def replace_with_liger_equivalents(root: torch.nn.Module) -> None:
    # Walk children first so we can safely replace in-place.
    for name, child in list(root.named_children()):
        replace_with_liger_equivalents(child)

        lig = getattr(child, "liger_equivalent", None)
        if lig is None:
            continue

        # Call the factory/method if it's callable; otherwise assume it's already a module.
        new_child = lig() if callable(lig) else lig

        # Keep training/eval state.
        try:
            new_child.train(child.training)
        except Exception:
            pass

        # Try to place on the same device/dtype as the original.
        ref = next(child.parameters(recurse=True), None)
        if ref is None:
            ref = next(child.buffers(recurse=True), None)
        if ref is not None:
            try:
                if ref.is_floating_point():
                    new_child.to(device=ref.device, dtype=ref.dtype)
                else:
                    new_child.to(device=ref.device)
            except Exception:
                pass
        rank0_print(f"Replaced {name} with liger equivalent")
        # Swap it in.
        setattr(root, name, new_child)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cosmos_rl.utils.parallelism import ParallelDims
import torch
from typing import Tuple


def convert_weight_from_hf(
    tensor: torch.Tensor,
    name: str,
    src_model_type: str,
    parallel_dims: ParallelDims,
    ignore_unknown_weights: bool = False,
) -> Tuple[str, torch.Tensor]:
    if parallel_dims.dp_shard_enabled or parallel_dims.cp_enabled:
        dp_shard_rank = parallel_dims.mesh[tuple(("dp_shard_cp",))].get_local_rank()
        dp_shard_size = parallel_dims.mesh[tuple(("dp_shard_cp",))].size()
    else:
        dp_shard_rank = 0
        dp_shard_size = 1

    dest_name = name
    shard = tensor
    # Do FSDP sharding
    shard = shard.contiguous()
    row_size = shard.shape[0]
    if row_size % dp_shard_size != 0:
        average_row_size = (row_size + dp_shard_size - 1) // dp_shard_size
        start_idx = dp_shard_rank * average_row_size
        end_idx = min(start_idx + average_row_size, row_size)
        shard = shard[start_idx:end_idx]
        return dest_name, shard
    else:
        shard = shard.tensor_split(dp_shard_size, dim=0)[dp_shard_rank]
        return dest_name, shard.contiguous()



import os
from typing import Optional, List, Callable, Union
from cosmos_rl.dispatcher.data.packer.base import DataPacker
from cosmos_rl.utils.logging import logger
from cosmos_rl.policy.config import Config as CosmosConfig
from torch.utils.data import Dataset


def main(
    dataset: Optional[Union[Dataset, Callable[[CosmosConfig], Dataset]]] = None,
    data_packer: Optional[DataPacker] = None,
    reward_fns: Optional[List[Callable]] = None,
    val_dataset: Optional[Dataset] = None,
    val_reward_fns: Optional[List[Callable]] = None,
    val_data_packer: Optional[DataPacker] = None,
    custom_logger_fns: Optional[List[Callable]] = None,
    *args,
    **kwargs,
):
    if kwargs:
        logger.warning(f"Unused kwargs: {list(kwargs.keys())}")

    role = os.environ.get("COSMOS_ROLE")
    assert role in ["Policy", "Rollout", "Controller"], f"Invalid role: {role}"
    if role == "Controller":
        from cosmos_rl.dispatcher.run_web_panel import main as controller_main

        controller_main(
            dataset=dataset,
            data_packer=data_packer,
            reward_fns=reward_fns,
            val_dataset=val_dataset,
            val_reward_fns=val_reward_fns,
            val_data_packer=val_data_packer,
            custom_logger_fns=custom_logger_fns,
        )
    elif role == "Policy":
        from cosmos_rl.policy.train import main as policy_main

        policy_main()
        return
    else:
        from cosmos_rl.rollout.rollout_entrance import run_rollout

        run_rollout()
        return


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Optional, Any, List, Dict
from torch.utils.data import Dataset, ConcatDataset
from datasets import load_dataset
from cosmos_rl.launcher.worker_entry import main as launch_worker
from cosmos_rl.policy.config import Config as CosmosConfig
from cosmos_rl.dispatcher.algo.reward import gsm8k_reward_fn
from transformers import AutoTokenizer
from cosmos_rl.dispatcher.data.packer import DecoderOnlyLLMDataPacker, DataPacker
from cosmos_rl.utils.modelscope import modelscope_load_dataset
from cosmos_rl.utils.logging import logger


class GSM8kDataset(Dataset):
    def setup(self, config: CosmosConfig, tokenizer: AutoTokenizer, *args, **kwargs):
        """
        This method is optional and get called by launcher after being mounted
        `config`: config;
        `tokenizer`: tokenizer;
        """
        self.config = config
        self.tokenizer = tokenizer
        modelscope_dataset_if_enabled = modelscope_load_dataset(
            config.train.train_policy.dataset.name, subset_name="main", split="train"
        )
        if modelscope_dataset_if_enabled is None:
            self.dataset = load_dataset("openai/gsm8k", "main", split="train")
        else:
            self.dataset = modelscope_dataset_if_enabled

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx: int) -> tuple[str, str]:
        """
        For DecoderOnlyLLMDataPacker, it should either return:
        - raw text prompt to be converted into input_ids by both rollout and policy models;
        - conversation format:
        ```
        [
            {
                "role": "user",
                "content": f"{question} Let\'s think step by step and output the final answer after \"####\".",
            }
        ]
        ```
        """
        assert hasattr(
            self, "tokenizer"
        ), "`self.tokenizer` should be set by the launcher"
        question = self.dataset[idx]["question"]
        assert isinstance(
            question, str
        ), f"Prompt should be a string, but got {type(question)}, {question}"
        # Convert to templated prompt
        conversation = [
            {
                "role": "user",
                "content": f'{question} Let\'s think step by step and output the final answer after "####".',
            }
        ]
        prompt = self.tokenizer.apply_chat_template(
            conversation,
            tokenize=False,
            add_generation_prompt=True,
        )
        return prompt

    def get_reference_answer(self, idx: int) -> Any:
        """
        This is mandatory for GRPO to get a reference answer for reward computation.
        """
        return self.dataset[idx]["answer"]


class GSM8kValDataset(GSM8kDataset):
    """
    This is a validation dataset for GSM8K, which is used to evaluate the performance of the model.
    It should be used in the launcher to evaluate the model during training.
    """

    def setup(self, config: CosmosConfig, tokenizer: AutoTokenizer, *args, **kwargs):
        if not config.train.enable_validation:
            logger.warning(
                "Validation is not enabled in the config. Skipping setup for GSM8kValDataset."
            )
            return

        self.config = config
        self.tokenizer = tokenizer

        self.dataset = load_dataset(
            config.validation.dataset.name, config.validation.dataset.subset
        )
        if config.validation.dataset.split:
            if isinstance(config.validation.dataset.split, list):
                dataset_list = []
                for split_name in config.validation.dataset.split:
                    dataset_list.append(self.dataset[split_name])
                self.dataset = ConcatDataset(dataset_list)
            else:
                assert isinstance(config.validation.dataset.split, str)
                self.dataset = self.dataset[config.validation.dataset.split]


def custom_reward_fn(
    to_be_evaluated: str, reference: Optional[Any] = None, *args, **kwargs
) -> float:
    assert isinstance(reference, str), "Reference answer should be a string"
    reward = gsm8k_reward_fn(to_be_evaluated, reference, *args, **kwargs)
    # Add more reward functions here
    # ...
    return reward


def custom_logger_fn(data: dict, step: int) -> None:
    """
    Example of custom logger function.
    Please mimic this to implement your own custom logger function.
    Your custom logger function is in the same format with this `custom_logger_fn`.
    It shoule be with two arguments: `data` and `step`.

    Arguments:
        The data argument will always be the follwing format:
            data = {
                "train/loss_avg": total_loss_avg,
                "train/loss_max": total_loss_max,
                "train/learning_rate": total_learning_rate,
                "train/iteration_time": total_iter_time_avg,
                "train/kl_loss_avg": total_kl_loss_avg,
                "train/kl_loss_max": total_kl_loss_max,
                "train/grad_norm": total_grad_norm,
            }
        The data argument includes the infomation could be logged.
        The step argument is the number for the current training step.

    Usage:
        After implementing this `custom_logger_fn` function, you can pass it through the `custom_logger_fns` argument of the `launch_worker`.
        The `custom_logger_fns` argument of the `launch_worker` is a list of Callable, which can includes multiple custom logger functions.
        Then the given custom logger functions will be called with the appropriate arguments during the training process.
    """
    # total_loss_avg = data.get("train/loss_avg", None)
    # total_loss_max = data.get("train/loss_max", None)
    # total_learning_rate = data.get("train/learning_rate", None)
    # total_iter_time_avg = data.get("train/iteration_time", None)
    # total_kl_loss_avg = data.get("train/kl_loss_avg", None)
    # total_kl_loss_max = data.get("train/kl_loss_max", None)
    # total_grad_norm = data.get("train/grad_norm", None)
    # This is only an example and no actual operations done for the logging.
    pass


class GSM8kDataPacker(DataPacker):
    """
    This is a demo data packer that wraps the underlying data packer of the selected model.
    This is meaningless for this example, but useful for explaining:
        - how dataset data is processed and collated into a mini-batch for rollout engine;
        - how rollout output is processed and collated into a mini-batch for policy model;
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Check source code of DecoderOnlyLLMDataPacker to see how it's implemented
        self.underlying_data_packer = DecoderOnlyLLMDataPacker()

    def setup(self, config: CosmosConfig, tokenizer: AutoTokenizer, *args, **kwargs):
        """
        This method is optional and get called by launcher after being mounted
        `config`: config;
        `tokenizer`: tokenizer;
        """
        super().setup(config, tokenizer, *args, **kwargs)
        self.underlying_data_packer.setup(config, tokenizer, *args, **kwargs)

    def get_rollout_input(self, item: Any) -> Any:
        """
        Convert dataset item into what rollout engine (e.g. vllm) expects
        """
        return self.underlying_data_packer.get_rollout_input(item)

    def rollout_collate_fn(self, items: List[Any]) -> Any:
        """
        Collate the rollout inputs into a mini-batch for rollout engine
        """
        return self.underlying_data_packer.rollout_collate_fn(items)

    def get_policy_input(
        self, item: Any, rollout_output: str, n_ignore_prefix_tokens: int = 0
    ) -> Any:
        """
        Process samples & rollout output before collating them into a mini-batch
        """
        return self.underlying_data_packer.get_policy_input(
            item, rollout_output, n_ignore_prefix_tokens
        )

    def policy_compute_max_len(self, processed_samples: List[Any]) -> int:
        """
        Compute the maximum sequence length of the mini-batch
        """
        return self.underlying_data_packer.policy_compute_max_len(processed_samples)

    def policy_collate_fn(
        self, processed_samples: List[Any], computed_max_len: int
    ) -> Dict[str, Any]:
        """
        Collate the mini-batch into the kwargs required by the policy model
        """
        return self.underlying_data_packer.policy_collate_fn(
            processed_samples, computed_max_len
        )


if __name__ == "__main__":

    def get_dataset(config: CosmosConfig) -> Dataset:
        return GSM8kDataset()

    def get_val_dataset(config: CosmosConfig) -> Dataset:
        return GSM8kValDataset()

    # It is best practice to pass the dataset as a factory function
    launch_worker(
        dataset=get_dataset,
        val_dataset=get_val_dataset,
        # Override the reward functions defined in toml
        reward_fns=[custom_reward_fn],
        # Optional: if not provided, the default data packer of the selected model will be used
        data_packer=GSM8kDataPacker(),
        val_data_packer=GSM8kDataPacker(),
        custom_logger_fns=[custom_logger_fn],
    )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import torch

from cosmos_rl.utils.parallelism import ParallelDims
from cosmos_rl.policy.config import Config as CosmosConfig
from cosmos_rl.comm.base import CommMixin
from cosmos_rl.dispatcher.protocol import Role
import cosmos_rl.utils.util as util
from transformers import AutoTokenizer


class State:
    UNINITIALIZED = 0
    WEIGHT_SYNCED = 1
    PROMPT_FETCH_END = 1 << 1
    PROMPT_CONSUME_END = 1 << 2

    _state: int = UNINITIALIZED

    def __init__(self):
        self._state = self.UNINITIALIZED

    def weight_synced(self):
        return (self._state & self.WEIGHT_SYNCED) != 0

    def set_weight_synced(self):
        self._state = self._state | self.WEIGHT_SYNCED

    def prompt_fetch_end(self):
        return (self._state & self.PROMPT_FETCH_END) != 0

    def set_prompt_fetch_end(self):
        self._state = self._state | self.PROMPT_FETCH_END

    def prompt_consume_end(self):
        return (self._state & self.PROMPT_CONSUME_END) != 0

    def set_prompt_consume_end(self):
        assert (
            not self.prompt_consume_end()
        ), "Prompt consume end event should not be set twice."
        self._state = self._state | self.PROMPT_CONSUME_END


class RolloutWorkerBase(CommMixin):
    def __init__(self, config: CosmosConfig, parallel_dims: ParallelDims) -> None:
        super().__init__()
        self.config = config
        self.role = Role.ROLLOUT
        self.parallel_dims = parallel_dims
        self.local_rank = int(os.environ.get("LOCAL_RANK", 0))  # rank in the node
        self.global_rank = int(os.environ.get("RANK", 0))  # rank in replica
        self.world_size = int(os.environ.get("WORLD_SIZE", 1))
        self.device = torch.device(f"cuda:{self.local_rank}")
        torch.cuda.set_device(self.device)

        self.tokenizer = util.retry(AutoTokenizer.from_pretrained)(
            self.config.policy.model_name_or_path
        )

        self.backend = "vllm"

        # Initialize the communication to controller.
        self.init_comm()
        self.init_redis()


class TRTLLMRolloutWorkerBase(CommMixin):
    """
    This class is special for TRTLLM, which will be compatible with PyExecutor in trtllm.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def post_init(
        self, cosmos_config: CosmosConfig, parallel_dims: ParallelDims, init_comm=True
    ):
        self.config = cosmos_config
        self.role = Role.ROLLOUT
        self.parallel_dims = parallel_dims
        self.local_rank = int(os.environ.get("LOCAL_RANK", 0))  # rank in the node
        self.global_rank = int(os.environ.get("RANK", 0))  # rank in replica
        self.world_size = int(os.environ.get("WORLD_SIZE", 1))
        self.device = torch.device(f"cuda:{self.local_rank}")
        torch.cuda.set_device(self.device)

        self.tokenizer = util.retry(AutoTokenizer.from_pretrained)(
            self.config.policy.model_name_or_path
        )
        self.backend = "trtllm"

        if init_comm:
            # Initialize the communication to controller.
            self.init_comm()
            self.init_redis()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import glob
import toml
import atexit
import subprocess
import torch

WORK_DIR = "/tmp/test_grad_allreduce"
CTRL_URL = ""


def write_train_config(dp_replica: int = 1, dp_shard: int = 1):
    # make sure every step has the same batch size
    train_batch_per_replica = 8 // dp_replica

    config = f"""
redis = "12808"

[train]
resume = false
epoch = 1
output_dir = "{WORK_DIR}"
epsilon = 1e-6
optm_name = "AdamW"
optm_lr = 1e-6
optm_impl = "fused"
optm_weight_decay = 0.01
optm_betas = [ 0.9, 0.999,]
optm_warmup_steps = 20
optm_grad_norm_clip = 1.0
async_tp_enabled = false
compile = false
param_dtype = "bfloat16"
fsdp_reduce_dtype = "float32"
fsdp_offload = false
fsdp_reshard_after_forward = "default"
train_batch_per_replica = {train_batch_per_replica}
sync_weight_interval = 1

[rollout]
gpu_memory_utilization = 0.7
enable_chunked_prefill = false
max_response_length = 2048
n_generation = 16
batch_size = 1
quantization = "none"


[policy]
model_name_or_path = "Qwen/Qwen2.5-3B-Instruct"
model_max_length = 4096
model_gradient_checkpointing = true

[logging]
logger = ['console', 'wandb']
project_name = "cosmos_rl"
experiment_name = "None"

[train.train_policy]
type = "grpo"
dataset.name = "JiaxinTsao/math_examples"
prompt_column_name = "prompt"
response_column_name = "result"
reward_function = "boxed_math"
dataset.split="train"
temperature = 1
epsilon_low = 0.2
epsilon_high = 0.2
kl_beta = 0.0

[train.ckpt]
enable_checkpoint = true
save_freq = 20
save_mode = "async"

[rollout.parallelism]
n_init_replicas = 1
tp_size = 2
pp_size = 1

[rollout.sampling_config]
temperature = 0.9
top_p = 1.0
top_k = 10

[policy.parallelism]
n_init_replicas = 1
tp_size = 2
cp_size = 1
dp_shard_size = {dp_shard}
pp_size = 1
dp_replicate_size = 1
"""
    cfg_file_path = os.path.join(WORK_DIR, "train_config.toml")
    with open(cfg_file_path, "w") as f:
        f.write(config)
    return cfg_file_path


def launch_controller(config: str, port: int = 8010):
    """
    Launch the controller process.
    """
    print("launch controller")
    p = subprocess.Popen(
        "python -m cosmos_rl.dispatcher.run_web_panel "
        f"--port {port} --config {config}",
        shell=True,
        stdout=sys.stdout,
        stderr=sys.stderr,
        text=True,
        env=dict(os.environ),
    )

    # get the controller listen port
    os.environ["COSMOS_CONTROLLER_HOST"] = f"localhost:{port}"
    global CTRL_URL
    CTRL_URL = os.environ["COSMOS_CONTROLLER_HOST"]
    return [p]


# launch_rollout will send training data to the controller in reproduciable sort
def launch_rollout(config: str):
    """
    Launch the rollout process.
    """
    print("launch rollout")

    cmd = f"CUDA_VISIBLE_DEVICES=6,7 COSMOS_CONTROLLER_HOST={CTRL_URL} ./tools/launch_replica.sh --type rollout --ngpus 2"
    print(f"launch rollout with cmd: {cmd}")
    return [
        subprocess.Popen(
            cmd, shell=True, stdout=sys.stdout, stderr=sys.stderr, text=True
        )
    ]


def launch_policy(config: str, gpus: str):
    """
    Launch the policy process.
    """
    print("launch policy")
    cfg = toml.load(config)
    policy_p = cfg["policy"]["parallelism"]
    ngpu = (
        int(policy_p["dp_shard_size"])
        * int(policy_p["tp_size"])
        * int(policy_p["pp_size"])
    )
    assert ngpu <= 4, "for test, policy will max use 4 gpus"

    cmd = f"CUDA_VISIBLE_DEVICES={gpus} COSMOS_CONTROLLER_HOST={CTRL_URL} ./tools/launch_replica.sh --type policy --ngpus {ngpu}"
    print(f"launch policy with cmd: {cmd}")
    return [
        subprocess.Popen(
            cmd, shell=True, stdout=sys.stdout, stderr=sys.stderr, text=True
        )
    ]


def compare_ckpt_grad(dp_shard_ckpt_path: str, dp_replica_ckpt_path: str):
    """
    Compare the gradients of the checkpointed model with the original model.
    """
    dp_shard_ckpt = torch.load(dp_shard_ckpt_path)
    dp_replica_ckpt = torch.load(dp_replica_ckpt_path)

    print("=========== Gradient Compare result: ===========")
    print(f" {dp_shard_ckpt_path}\n {dp_replica_ckpt_path}")
    for key in dp_shard_ckpt.keys():
        if key not in dp_replica_ckpt:
            print(f"Key {key} not found in dp_replica_ckpt")
            continue
        # for dtype bfloat16, we release tolerance to 1e-3
        if not torch.allclose(
            dp_shard_ckpt[key], dp_replica_ckpt[key], atol=1e-3, rtol=1e-3
        ):
            print(f"Gradients do not match for key {key}")
            return False
    print("All gradients match!")
    return True


popens = []


def cleanup():
    """
    Cleanup function to kill all processes.
    """
    for popen in popens:
        popen.kill()
    popens.clear()

    # try kill created processes
    try:
        login_user = os.getlogin()
    except Exception:
        login_user = "root"

    subprocess.run(f"pkill -u {login_user} -f 'cosmos'", shell=True)
    subprocess.run(f"pkill -u {login_user} -f 'redis'", shell=True)
    subprocess.run(f"pkill -u {login_user} -f 'torchrun'", shell=True)


atexit.register(cleanup)


if __name__ == "__main__":
    # Set the environment variable for the test
    os.environ["COSMOS_TEST"] = "1"
    os.environ["COSMOS_LOG_LEVEL"] = "INFO"

    # Create the work directory
    os.makedirs(WORK_DIR, exist_ok=True)

    # # step 1: run a normal training with dp_replica=1, dp_shard=2
    # print("step 1: run a normal training with dp_replica=1, dp_shard=2")
    # # Write the training config file
    # config = write_train_config(dp_replica=1, dp_shard=2)
    # # Launch the controller
    # popens += launch_controller(config)
    # # Launch the rollout
    # popens += launch_rollout(config)
    # # Launch the policy process
    # popens += launch_policy(config, "0,1,2,3")
    # # wait for policy finish
    # popens[-1].wait()
    # # kill all processes
    # cleanup()

    # # wait for a while to ensure all processes are killed
    # time.sleep(2)

    # step 2: run a normal training with dp_replica=2, dp_shard=1
    print("step 2: run a normal training with dp_replica=2, dp_shard=1")
    # Write the training config file
    config = write_train_config(dp_replica=2, dp_shard=1)
    # Launch the controller
    popens += launch_controller(config, 8300)
    # Mock rollout
    popens += launch_rollout(config)
    # Launch 2 policy process
    popens += launch_policy(config, "0,1")
    popens += launch_policy(config, "2,3")

    # wait for policy finish
    popens[-1].wait()
    # kill all processes
    cleanup()

    # step 3: compare the gradients of the checkpointed model with the original model
    print(
        "step 3: compare the gradients of the checkpointed model with the original model"
    )
    dumped_grad = glob.glob(os.path.join(WORK_DIR, "grad_embed_tokens.weight-*.pt"))
    dp1r2 = [fn for fn in dumped_grad if "1-2" in os.path.basename(fn)][0]
    dp2r1 = [fn for fn in dumped_grad if "2-1" in os.path.basename(fn)][0]

    compare_ckpt_grad(
        dp1r2,
        dp2r1,
    )

    # step 4: clean file
    for fn in dumped_grad:
        os.remove(fn)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import socket
import sys

from . import replica
from . import profiler
from . import nccl
from . import algo
from .utils import console


@click.group()
def cosmos():
    """
    Cosmos Reason1 CLI.
    """
    pass


replica.add_command(cosmos)
profiler.add_command(cosmos)
nccl.add_command(cosmos)
algo.add_command(cosmos)


def get_ip_from_hostname(hostname):
    try:
        ip_address = socket.gethostbyname(hostname)
    except socket.error as e:
        console.print(f"Failed to get IP from hostname: {e}")
        sys.exit(1)
    return ip_address


if __name__ == "__main__":
    cosmos()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import requests
import sys


from functools import partial
from urllib.parse import urljoin

from cosmos_rl.utils.network_util import make_request_with_retry
from cosmos_rl.cli.utils import console, get_base_url
from cosmos_rl.utils.api_suffix import COSMOS_API_SET_PROFILE_SUFFIX
from cosmos_rl.cli.custom_group import ControllerNeedeGroup


@click.group(cls=ControllerNeedeGroup)
def profile():
    """
    Manage profiler behavior of the training.
    """
    pass


def comma_separated_int_list(ctx, param, value):
    if not value:
        return []
    return [int(v.strip()) for v in value.split(",")]


@profile.command(name="set")
@click.argument("replica_name", type=str)
@click.option(
    "--active-steps", "-as", type=int, help="The number of steps that profiler traces."
)
@click.option(
    "--rank-filter",
    "-rf",
    callback=comma_separated_int_list,
    help="The ranks that profiler traces, separated by comma.",
)
@click.option(
    "--record-shape",
    "-rs",
    is_flag=True,
    help="Whether to record the shape of the tensor in the profiler.",
)
@click.option(
    "--profile-memory",
    "-pm",
    is_flag=True,
    help="Whether to profile the memory of the tensor in the profiler.",
)
@click.option(
    "--with-stack",
    "-ws",
    is_flag=True,
    help="Whether to record the stack trace of the profiler.",
)
@click.option(
    "--with-modules",
    "-wm",
    is_flag=True,
    help="Whether to record the modules of the profiler.",
)
def set_profile(
    replica_name,
    active_steps,
    rank_filter,
    record_shape,
    profile_memory,
    with_stack,
    with_modules,
    controller_host: str,
    controller_port: str,
):
    """
    Set profile mode of the replica named REPLICA_NAME.
    """
    base_url = get_base_url(controller_host, controller_port)
    url = urljoin(base_url, COSMOS_API_SET_PROFILE_SUFFIX)
    if len(rank_filter) == 0:
        rank_filter = [
            0,
        ]
    if active_steps is None:
        active_steps = 1
    response = make_request_with_retry(
        partial(
            requests.post,
            json={
                "replica_name": replica_name,
                "active_steps": active_steps,
                "rank_filter": rank_filter,
                "record_shape": record_shape,
                "profile_memory": profile_memory,
                "with_stack": with_stack,
                "with_modules": with_modules,
            },
        ),
        [url],
        max_retries=4,
    )

    if response.status_code != 200:
        console.print(f"Failed to set profile for replica: {replica_name}")
        sys.exit(1)

    msg = response.json()["message"]
    console.print(msg)


def add_command(cli_group):
    cli_group.add_command(profile)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from flash_attn import (
    flash_attn_func as ori_flash_attn_func,
    flash_attn_varlen_func as ori_flash_attn_varlen_func,
)
from flash_attn.layers.rotary import apply_rotary_emb as ori_apply_rotary_emb


def _flash_attn_func(*args, **kwargs):
    return ori_flash_attn_func(*args, **kwargs)


def _flash_attn_varlen_func(*args, **kwargs):
    return ori_flash_attn_varlen_func(*args, **kwargs)


def _apply_rotary_emb(*args, **kwargs):
    return ori_apply_rotary_emb(*args, **kwargs)


flash_attn_func = _flash_attn_func
flash_attn_varlen_func = _flash_attn_varlen_func
apply_rotary_emb = _apply_rotary_emb


def set_flash_attn_deterministic(deterministic: bool):
    global flash_attn_func, flash_attn_varlen_func, apply_rotary_emb
    if deterministic:
        flash_attn_func = partial(_flash_attn_func, deterministic=True)
        flash_attn_varlen_func = partial(_flash_attn_varlen_func, deterministic=True)
    else:
        flash_attn_func = _flash_attn_func
        flash_attn_varlen_func = _flash_attn_varlen_func



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#!/usr/bin/env python3

import socket
import subprocess
import sys
import logging
import time
import os
import re
import argparse
from typing import List, Dict, Optional, Any, Callable
import toml
import tempfile

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("cosmos")


# ---------------------------------------------------------------------------
# Queue priority helper
# ---------------------------------------------------------------------------
# Map numeric queue-priority (1-9) to the corresponding Lepton priority_class
# string expected by the backend API.
#
# 1-3  → low-1000 / 2000 / 3000
# 4-6  → mid-4000 / 5000 / 6000
# 7-9  → high-7000 / 8000 / 9000
#
# Note: keep in sync with lepton-cli definitions.
NUM_PRIORITY_MAPPING = {
    1: "low-1000",
    2: "low-2000",
    3: "low-3000",
    4: "mid-4000",
    5: "mid-5000",
    6: "mid-6000",
    7: "high-7000",
    8: "high-8000",
    9: "high-9000",
}


def wait_for_url_ready(url: str, process: Optional[subprocess.Popen] = None):
    """
    Wait for a URL to be ready by sending a GET request.

    Args:
        url: The URL to check

    Returns:
        None
    """
    while True:
        # create TCP socket
        try:
            if process is not None:
                if process.poll() is not None:
                    if process.returncode != 0:
                        logger.error(
                            f"Process {process.pid} exited with code {process.returncode}. Exiting."
                        )
                        sys.exit(1)
                    else:
                        logger.error(
                            f"Process {process.pid} exited as soon as launched. Exiting."
                        )
                        sys.exit(1)
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(1)
            host, port = url.split(":")
            sock.connect((host, int(port)))
            sock.close()
            break
        except socket.error:
            # If the connection fails, wait and retry
            time.sleep(1)


def read_config(config_file: str) -> Dict[str, Any]:
    """
    Read configuration from a TOML file.

    Args:
        config_file: Path to the TOML configuration file

    Returns:
        Dictionary containing the configuration
    """
    try:
        with open(config_file, "r") as f:
            config = toml.load(f)
        return config
    except Exception as e:
        logger.error(f"Error reading config file {config_file}: {e}")
        sys.exit(1)


def get_available_gpus() -> List[str]:
    """
    Detect available GPUs using nvidia-smi and return their IDs.

    Returns:
        List of GPU IDs as strings
    """
    try:
        cmd = ["nvidia-smi", "--query-gpu=index", "--format=csv,noheader"]
        cvd = os.getenv("CUDA_VISIBLE_DEVICES", None)
        if cvd is not None:
            # Add the GPU IDs to the command
            cmd += ["--id=" + cvd]
        # Run nvidia-smi to get GPU information
        result = subprocess.run(cmd, capture_output=True, text=True, check=True)

        # Parse the output to get GPU IDs
        gpu_ids = [line.strip() for line in result.stdout.splitlines()]

        if not gpu_ids:
            logger.error("Warning: No GPUs detected")
            return []

        logger.info(f"Detected {len(gpu_ids)} GPUs: {', '.join(gpu_ids)}")
        return gpu_ids

    except subprocess.CalledProcessError as e:
        logger.error(f"Error running nvidia-smi: {e}")
        return []
    except Exception as e:
        logger.error(f"Error detecting GPUs: {e}")
        return []


def launch_processes(
    commands: List[str],
    gpu_devices: Optional[List[str]],
    control_urls: Optional[List[str]],
    output_files: Optional[List[str]],
    extra_env: Optional[Dict[str, str]] = None,
) -> List[subprocess.Popen]:
    """
    Launch multiple subprocesses and return their process objects.

    Args:
        commands: List of command strings to execute
        gpu_devices: List of GPU device IDs to assign to each process (e.g., ["0", "1", "2"])
        control_urls: List of controller URLs to assign to each process (e.g., ["localhost:8000"])
        output_files: List of output files to redirect process output to (e.g., ["output1.log", "output2.log"])

    Returns:
        List of Popen objects for the launched processes
    """
    processes = []

    if gpu_devices is None:
        gpu_devices = [None] * len(commands)
    elif len(gpu_devices) != len(commands):
        raise ValueError("Number of GPU devices must match number of commands")

    for cmd, gpu_id, url, ofile in zip(
        commands, gpu_devices, control_urls, output_files
    ):
        try:
            # Prepare environment variables
            env = dict(os.environ)
            if gpu_id is not None:
                env["CUDA_VISIBLE_DEVICES"] = gpu_id
            if url is not None:
                env["COSMOS_CONTROLLER_HOST"] = url
            if extra_env is not None:
                env.update(extra_env)
            if ofile is not None:
                f = open(ofile, "wb")
                cout = f
                cerr = f
            else:
                cout = sys.stdout
                cerr = sys.stderr

            # Launch process and capture output
            logger.info(f"Launching process with command: {cmd}")
            process = subprocess.Popen(
                cmd, shell=True, stdout=cout, stderr=cerr, env=env
            )
            processes.append(process)
            if ofile is not None:
                f.close()
        except Exception as e:
            logger.error(f"Error launching process for command '{cmd}': {e}")

    return processes


def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Launch multiple processes with GPU assignments"
    )
    parser.add_argument(
        "--config",
        type=str,
        required=True,
        help="Path to TOML configuration file, which specifies the detailed configuration for the whole training process including algorithm, model, data, parallelism, etc.",
    )
    parser.add_argument(
        "--url",
        type=str,
        default=None,
        help="URL of the controller for the policy and rollout replicas to connect to, consisting of IP and port in the format ip:port. If not provided, the controller will be launched on the local machine. If provided and the IP is the local IP, the controller will be launched on the local machine. If provided and the IP is not the local IP, the controller will be launched on the remote machine.",
    )
    parser.add_argument(
        "--port",
        type=int,
        default=8000,
        help="Port of the controller to connect to, default is 8000. This is only used when --url is not provided to launch the controller on the local machine.",
    )
    parser.add_argument(
        "--policy",
        type=int,
        default=None,
        help="Total number of policy replicas to launch in the whole system. If not provided, the number of policy replicas will be obtained from TOML configuration file.",
    )
    parser.add_argument(
        "--rollout",
        type=int,
        default=None,
        help="Total number of rollout replicas to launch in the whole system. If not provided, the number of rollout replicas will be obtained from TOML configuration file.",
    )
    parser.add_argument(
        "--p2r-ratio",
        type=str,
        default=None,
        help="Ratio of policy replicas to rollout replicas. This is used to determine the number of rollout replicas and the number of policy replicas based on the number of workers.",
    )

    parser.add_argument(
        "--log-dir",
        type=str,
        default=None,
        help="Directory to save logs. If not provided, logs will be printed to stdout.",
    )
    parser.add_argument(
        "--num-workers",
        type=int,
        default=None,
        help="Number of workers to use for the job, default is 1. This is used when multi-node training are used for the job.",
    )
    parser.add_argument(
        "--worker-idx",
        type=int,
        default=0,
        help="Worker index for local execution. In Lepton mode, this is ignored as worker indices are automatically assigned by Lepton.",
    )

    parser.add_argument(
        "--node-ip-list",
        type=str,
        default=None,
        help="list of ips for all the workers, separated by ';'. This is used when multi-node training are used for one replica.",
    )

    parser.add_argument(
        "--rdzv-port",
        type=int,
        default=29345,
        help="Rendezvous endpoint port for the job, default is 29345. This is used when multi-node training are used for one replica.",
    )

    parser.add_argument(
        "script",
        nargs="?",  # “?” means 0 or 1 occurrences
        default=None,
        help="A user script which can be provided for custom dataset, reward functions, and model registration.",
    )

    parser.add_argument(
        "--lepton-mode",
        action="store_true",
        default=False,
        help="Enable Lepton mode for remote execution",
    )

    parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help="Enable log level to debug",
    )

    # Lepton specific options
    lepton_group = parser.add_argument_group("Lepton mode options")
    lepton_group.add_argument("--lepton-job-name", "-n", type=str, help="Job name")
    lepton_group.add_argument(
        "--lepton-container-image", type=str, help="Container image for the job"
    )
    lepton_group.add_argument(
        "--lepton-container-port",
        type=str,
        help="Ports to expose for the job, in the format portnumber[:protocol]",
        action="append",
    )
    lepton_group.add_argument(
        "--lepton-resource-shape", type=str, help="Resource shape for the pod"
    )
    lepton_group.add_argument(
        "--lepton-node-group",
        "-ng",
        type=str,
        help="Node group for the job",
        action="append",
    )
    lepton_group.add_argument(
        "--lepton-max-failure-retry",
        type=int,
        help="Maximum number of failures to retry per worker",
    )
    lepton_group.add_argument(
        "--lepton-max-job-failure-retry",
        type=int,
        help="Maximum number of failures to retry per whole job",
    )
    lepton_group.add_argument(
        "--lepton-env",
        "-e",
        type=str,
        help="Environment variables to pass to the job, in the format `NAME=VALUE`",
        action="append",
    )
    lepton_group.add_argument(
        "--lepton-secret",
        "-s",
        type=str,
        help="Secrets to pass to the job",
        action="append",
    )
    lepton_group.add_argument(
        "--lepton-mount",
        type=str,
        help="Persistent storage to be mounted to the job",
        action="append",
    )
    lepton_group.add_argument(
        "--lepton-image-pull-secrets",
        type=str,
        help="Secrets to use for pulling images",
        action="append",
    )
    lepton_group.add_argument(
        "--lepton-intra-job-communication",
        type=bool,
        help="Enable intra-job communication",
    )
    lepton_group.add_argument(
        "--lepton-privileged",
        action="store_true",
        help="Run the job in privileged mode",
    )
    lepton_group.add_argument(
        "--lepton-ttl-seconds-after-finished",
        type=int,
        help="TTL for finished jobs in seconds",
        default=259200,
    )
    lepton_group.add_argument(
        "--lepton-log-collection",
        "-lg",
        type=bool,
        help="Enable or disable log collection",
    )
    lepton_group.add_argument(
        "--lepton-node-id", "-ni", type=str, help="Node for the job", action="append"
    )
    lepton_group.add_argument(
        "--lepton-queue-priority",
        "-qp",
        type=int,
        choices=list(NUM_PRIORITY_MAPPING.keys()),
        help=(
            "Queue priority for dedicated node groups. Provide a number 1-9 which"
            " will be mapped to priority classes low-1000 … high-9000."
        ),
    )
    # Whether the job can be preempted by higher-priority jobs (only valid for
    # dedicated node groups). Tri-state: flag present → True; absent → None.
    lepton_group.add_argument(
        "--lepton-can-be-preempted",
        "-cbp",
        action="store_true",
        default=None,
        help=(
            "Allow this job to be preempted by higher priority jobs (only for"
            " dedicated node groups)."
        ),
    )

    # Whether the job itself is allowed to preempt lower-priority jobs.
    lepton_group.add_argument(
        "--lepton-can-preempt",
        "-cp",
        action="store_true",
        default=None,
        help=(
            "Allow this job to preempt lower priority jobs (only for dedicated"
            " node groups)."
        ),
    )
    lepton_group.add_argument(
        "--lepton-visibility", type=str, help="Visibility of the job (public/private)"
    )
    lepton_group.add_argument(
        "--lepton-shared-memory-size", type=int, help="Shared memory size in MiB"
    )
    lepton_group.add_argument(
        "--lepton-with-reservation",
        type=str,
        help="Reservation ID for dedicated node groups",
    )

    args = parser.parse_args()

    # Validate Lepton mode arguments
    if args.lepton_mode:
        required_args = [("lepton_job_name", "--lepton-job-name")]

        for arg_name, arg_flag in required_args:
            if not getattr(args, arg_name):
                parser.error(f"{arg_flag} is required when --lepton-mode is enabled")

    return args, parser


def get_local_ip():
    """
    Get the local IP address of the machine.

    Returns:
        Local IP address as a string
    """
    try:
        import socket

        hostname = socket.gethostname()
        local_ip = socket.gethostbyname(hostname)
        return [local_ip, hostname]
    except Exception as e:
        logger.error(f"Error getting local IP address: {e}")
        return None


def replica_placement(
    available_gpus: List[int],
    n_policy: int,
    n_rollouts: int,
    min_n_gpus_policy: int,
    min_n_gpus_rollout: int,
    replica_script: str,
    control_url: str,
    output_dir: Optional[str],
    get_worker_ip: Optional[Callable] = None,
    rdzv_port: Optional[int] = None,
    script: Optional[str] = None,
    backend: str = "vllm",
    config_path: Optional[str] = None,
) -> List[List[str]]:
    commands = []
    gpu_devices = []
    control_urls = []
    output_files = []
    assert len(available_gpus) in [
        1,
        2,
        4,
        8,
    ], "Number of GPUs per worker must be 1, 2, 4, or 8"
    # Prepare the command to launch the controller for all workers
    global_available_gpus = [available_gpus]
    # Create commands for policy and rollout replicas
    gpu_idx = 0
    global_worker_idx = 0
    global_launch_settings = []
    # Assign launch settings for each worker
    for i in range(n_policy):
        if min_n_gpus_policy > len(global_available_gpus[global_worker_idx]):
            assert (
                min_n_gpus_policy % len(global_available_gpus[global_worker_idx]) == 0
            ), f"min_n_gpus_policy {min_n_gpus_policy} is not divisible by {len(global_available_gpus[global_worker_idx])}"
            nodes_needed = min_n_gpus_policy // len(
                global_available_gpus[global_worker_idx]
            )
            rdzv_ip = "localhost"
            for node_in_replica in range(nodes_needed):
                gpu_devices.append(
                    ",".join([str(g) for g in global_available_gpus[global_worker_idx]])
                )
                commands.append(
                    f"{replica_script} --type policy --ngpus {len(global_available_gpus[global_worker_idx])} --nnodes {nodes_needed} --backend {backend} --config {config_path}"
                )
                if script is not None:
                    commands[-1] += f" --script {script}"
                if node_in_replica == 0:
                    commands[-1] += f" --rdzv-endpoint {rdzv_ip}:{rdzv_port}"
                    if get_worker_ip is not None:
                        rdzv_ip = get_worker_ip(global_worker_idx)
                else:
                    commands[-1] += f" --rdzv-endpoint {rdzv_ip}:{rdzv_port}"

                control_urls.append(control_url)
                output_files.append(
                    os.path.join(output_dir, f"policy_{i}.log")
                    if output_dir is not None
                    else None
                )
                global_launch_settings.append(
                    [commands, gpu_devices, control_urls, output_files]
                )
                commands = []
                gpu_devices = []
                control_urls = []
                output_files = []
                global_worker_idx += 1
                global_available_gpus.append(available_gpus)
        else:
            if gpu_idx + min_n_gpus_policy > len(
                global_available_gpus[global_worker_idx]
            ):
                global_launch_settings.append(
                    [commands, gpu_devices, control_urls, output_files]
                )
                commands = []
                gpu_devices = []
                control_urls = []
                output_files = []
                gpu_idx = 0
                global_worker_idx += 1
                global_available_gpus.append(available_gpus)

            gpu_devices.append(
                ",".join(
                    [
                        str(g)
                        for g in global_available_gpus[global_worker_idx][
                            gpu_idx : gpu_idx + min_n_gpus_policy
                        ]
                    ]
                )
            )
            commands.append(
                f"{replica_script} --type policy --ngpus {min_n_gpus_policy} --backend {backend} --config {config_path}"
            )
            if script is not None:
                commands[-1] += f" --script {script}"
            control_urls.append(control_url)
            output_files.append(
                os.path.join(output_dir, f"policy_{i}.log")
                if output_dir is not None
                else None
            )
            gpu_idx += min_n_gpus_policy

    if min_n_gpus_rollout > len(global_available_gpus[global_worker_idx]):
        # If the number of GPUs needed for rollout is more than available GPUs, we need to allocate a new worker
        if gpu_idx > 0:
            global_launch_settings.append(
                [commands, gpu_devices, control_urls, output_files]
            )
            commands = []
            gpu_devices = []
            control_urls = []
            output_files = []
            gpu_idx = 0
            global_worker_idx += 1
            global_available_gpus.append(available_gpus)

    for i in range(n_rollouts):
        if min_n_gpus_rollout > len(global_available_gpus[global_worker_idx]):
            assert (
                min_n_gpus_rollout % len(global_available_gpus[global_worker_idx]) == 0
            ), f"min_n_gpus_rollout {min_n_gpus_rollout} is not divisible by {len(global_available_gpus[global_worker_idx])}"
            nodes_needed = min_n_gpus_rollout // len(
                global_available_gpus[global_worker_idx]
            )
            rdzv_ip = "localhost"
            for node_in_replica in range(nodes_needed):
                gpu_devices.append(
                    ",".join([str(g) for g in global_available_gpus[global_worker_idx]])
                )
                commands.append(
                    f"{replica_script} --type rollout --ngpus {len(global_available_gpus[global_worker_idx])} --nnodes {nodes_needed} --backend {backend} --config {config_path}"
                )
                if script is not None:
                    commands[-1] += f" --script {script}"
                if node_in_replica == 0:
                    commands[-1] += f" --rdzv-endpoint {rdzv_ip}:{rdzv_port}"
                    if get_worker_ip is not None:
                        rdzv_ip = get_worker_ip(global_worker_idx)
                else:
                    commands[-1] += f" --rdzv-endpoint {rdzv_ip}:{rdzv_port}"

                control_urls.append(control_url)
                output_files.append(
                    os.path.join(output_dir, f"rollout_{i}.log")
                    if output_dir is not None
                    else None
                )
                global_launch_settings.append(
                    [commands, gpu_devices, control_urls, output_files]
                )
                commands = []
                gpu_devices = []
                control_urls = []
                output_files = []
                global_worker_idx += 1
                global_available_gpus.append(available_gpus)
        else:
            if gpu_idx + min_n_gpus_rollout > len(
                global_available_gpus[global_worker_idx]
            ):
                global_launch_settings.append(
                    [commands, gpu_devices, control_urls, output_files]
                )
                commands = []
                gpu_devices = []
                control_urls = []
                output_files = []
                gpu_idx = 0
                global_worker_idx += 1
                global_available_gpus.append(available_gpus)

            gpu_devices.append(
                ",".join(
                    [
                        str(g)
                        for g in available_gpus[gpu_idx : gpu_idx + min_n_gpus_rollout]
                    ]
                )
            )
            commands.append(
                f"{replica_script} --type rollout --ngpus {min_n_gpus_rollout} --backend {backend} --config {config_path}"
            )
            if script is not None:
                commands[-1] += f" --script {script}"
            control_urls.append(control_url)
            output_files.append(
                os.path.join(output_dir, f"rollout_{i}.log")
                if output_dir is not None
                else None
            )
            gpu_idx += min_n_gpus_rollout

    if len(commands) > 0:
        global_launch_settings.append(
            [commands, gpu_devices, control_urls, output_files]
        )
    return global_launch_settings


def main():
    args, parser = parse_args()
    if args.debug:
        os.environ["COSMOS_LOG_LEVEL"] = "DEBUG"

    # Check if the config file is provided
    cosmos_config = read_config(args.config)
    if args.script is not None and args.script.endswith(".py"):
        # If the script is a Python file, we need to make sure it is absolute path
        # so that it can be found by the launched processes
        script = os.path.abspath(args.script)
    else:
        script = args.script if args.script is not None else None

    # Get the number of GPUs required for policy and rollout
    # and the number of replicas for each
    policy_parallelism = cosmos_config.get("policy", {}).get("parallelism", {})
    rollout_parallelism = cosmos_config.get("rollout", {}).get("parallelism", {})
    # Calculate the minimum number of GPUs required for policy and rollout
    # based on the parallelism settings in the configuration
    # Treat dp_shard_size as 1 if it is not set
    min_n_gpus_policy = (
        policy_parallelism.get("tp_size", 1)
        * policy_parallelism.get("dp_replicate_size", 1)
        * policy_parallelism.get("pp_size", 1)
        * policy_parallelism.get("cp_size", 1)
    )
    min_n_gpus_rollout = (
        rollout_parallelism.get("tp_size", 1)
        * rollout_parallelism.get("dp_replicate_size", 1)
        * rollout_parallelism.get("pp_size", 1)
        * rollout_parallelism.get("cp_size", 1)
    )
    if policy_parallelism.get("dp_shard_size", 1) >= 1:
        min_n_gpus_policy = min_n_gpus_policy * policy_parallelism.get(
            "dp_shard_size", 1
        )
    if rollout_parallelism.get("dp_shard_size", 1) >= 1:
        min_n_gpus_rollout = min_n_gpus_rollout * rollout_parallelism.get(
            "dp_shard_size", 1
        )
    backend = cosmos_config.get("rollout", {}).get("backend", "vllm")
    logger.info(f"Using rollout backend: {backend}")

    if args.p2r_ratio is not None:
        assert (
            args.num_workers is not None
        ), "When using --p2r-ratio, --num-workers must be specified"
        p2r_ratio = args.p2r_ratio.split(":")
        assert (
            len(p2r_ratio) == 2
        ), "Invalid --p2r-ratio format. Use 'policy:rollout' format."
        p_ratio = int(p2r_ratio[0])
        r_ratio = int(p2r_ratio[1])

        if args.lepton_mode:
            match = re.search(r"(8|4|2)x", args.lepton_resource_shape)
            if match:
                num_gpus_per_node = int(match.group(1))
            else:
                num_gpus_per_node = 1
        else:
            num_gpus_per_node = len(get_available_gpus())

        num_per_ratio = (
            args.num_workers
            * num_gpus_per_node
            / (p_ratio * min_n_gpus_policy + r_ratio * min_n_gpus_rollout)
        )
        args.policy = int(num_per_ratio * p_ratio)
        args.rollout = int(num_per_ratio * r_ratio)
        assert args.policy >= 1, "Number of policy replicas must be at least 1"
        assert (
            args.policy * min_n_gpus_policy + args.rollout * min_n_gpus_rollout
            <= args.num_workers * num_gpus_per_node
        )

    if args.policy is None:
        n_policy = policy_parallelism.get("n_init_replicas", 1)
    else:
        n_policy = args.policy
    if args.rollout is None:
        n_rollouts = rollout_parallelism.get("n_init_replicas", 1)
    else:
        n_rollouts = args.rollout

    # If the training type is SFT, set n_rollouts to 0
    if (
        cosmos_config.get("train", {}).get("train_policy", {}).get("type", "grpo")
        == "sft"
    ):
        n_rollouts = 0

    # Handle Lepton mode
    if args.lepton_mode:
        from leptonai.api.v2.client import APIClient
        from leptonai.config import BASE_IMAGE, VALID_SHAPES
        from leptonai.api.v1.types.job import (
            LeptonJob,
            LeptonJobUserSpec,
            LeptonResourceAffinity,
            ReservationConfig,
        )
        from leptonai.api.v1.types.deployment import (
            LeptonLog,
            QueueConfig,
        )
        from leptonai.api.v1.types.common import Metadata, LeptonVisibility
        from leptonai.api.v1.photon import (
            make_env_vars_from_strings,
            make_mounts_from_strings,
        )
        from leptonai.cli.util import (
            _get_valid_nodegroup_ids,
            _get_valid_node_ids,
            make_container_port_from_string,
        )

        # Initialize Lepton client
        client = APIClient()

        # Create job specification
        job_spec = LeptonJobUserSpec()

        # Construct the original launch_processes command
        # Update policy and rollout numbers in the lepton config
        if "policy" in cosmos_config and "parallelism" in cosmos_config["policy"]:
            cosmos_config["policy"]["parallelism"]["n_init_replicas"] = n_policy
        if "rollout" in cosmos_config and "parallelism" in cosmos_config["rollout"]:
            cosmos_config["rollout"]["parallelism"]["n_init_replicas"] = n_rollouts
        config_content = toml.dumps(cosmos_config)
        launch_cmd = f"""\
cat >config.toml <<EOF
{config_content}
EOF

cosmos-rl --config config.toml"""

        # Get all non-Lepton arguments
        non_lepton_args = []
        for action in parser._actions:
            if hasattr(action, "option_strings") and action.option_strings:
                # Skip help action, lepton related arguments, and worker-idx
                if (
                    action.dest == "help"
                    or any(
                        opt.startswith("--lepton-") or opt == "--lepton-mode"
                        for opt in action.option_strings
                    )
                    or action.dest == "worker_idx"
                    or action.dest == "config"
                ):  # skip worker-idx
                    continue

                value = getattr(args, action.dest)
                if value is not None:
                    if isinstance(value, bool):
                        if value:
                            non_lepton_args.append(action.option_strings[0])
                    else:
                        non_lepton_args.append(f"{action.option_strings[0]} {value}")

        # Add all non-Lepton arguments to the command
        launch_cmd += " " + " ".join(non_lepton_args)
        if script is not None:
            launch_cmd += f" {script}"

        # Handle node groups, queue priority and preemption flags
        if (
            args.lepton_node_group
            or args.lepton_queue_priority is not None
            or args.lepton_can_be_preempted is not None
            or args.lepton_can_preempt is not None
        ):
            if (
                args.lepton_queue_priority is not None
                or args.lepton_can_be_preempted is not None
                or args.lepton_can_preempt is not None
            ) and not args.lepton_node_group:
                logger.error(
                    "Error: Queue priority is only available for dedicated node groups"
                )
                logger.error(
                    "Please use --lepton-queue-priority with --lepton-node-group"
                )
                sys.exit(1)

            node_group_ids = _get_valid_nodegroup_ids(
                args.lepton_node_group,
                need_queue_priority=(
                    args.lepton_queue_priority is not None
                    or args.lepton_can_be_preempted is not None
                    or args.lepton_can_preempt is not None
                ),
            )
            valid_node_ids = (
                _get_valid_node_ids(node_group_ids, args.lepton_node_id)
                if args.lepton_node_id
                else None
            )

            job_spec.affinity = LeptonResourceAffinity(
                allowed_dedicated_node_groups=node_group_ids,
                allowed_nodes_in_node_group=valid_node_ids,
            )

            if (
                args.lepton_queue_priority is not None
                or args.lepton_can_be_preempted is not None
                or args.lepton_can_preempt is not None
            ):
                # Ensure queue_config exists
                if job_spec.queue_config is None:
                    job_spec.queue_config = QueueConfig()

                priority_class = None
                if args.lepton_queue_priority is not None:
                    # Convert numeric priority to the Lepton priority_class string.
                    priority_class = NUM_PRIORITY_MAPPING[args.lepton_queue_priority]

                job_spec.queue_config.priority_class = priority_class or "mid-4000"

                if args.lepton_can_be_preempted is not None:
                    job_spec.queue_config.can_be_preempted = bool(
                        args.lepton_can_be_preempted
                    )

                if args.lepton_can_preempt is not None:
                    job_spec.queue_config.can_preempt = bool(args.lepton_can_preempt)

        # Set resource shape
        if args.lepton_resource_shape:
            job_spec.resource_shape = args.lepton_resource_shape
        else:
            available_types = "\n      ".join(VALID_SHAPES)
            logger.error(
                "Error: Missing option '--lepton-resource-shape'.\n"
                f"Available types are:\n      {available_types}.\n"
            )
            sys.exit(1)

        # Only for calculating the number of nodes needed.
        # Use this to replace the args.num_workers to calculate the number of nodes needed rather than specifying it using --num-workers
        # example

        match = re.search(r"(8|4|2)x", args.lepton_resource_shape)
        if match:
            num_gpus_per_node = int(match.group(1))
        else:
            num_gpus_per_node = 1
        global_launch_settings = replica_placement(
            list(range(num_gpus_per_node)),
            n_policy,
            n_rollouts,
            min_n_gpus_policy,
            min_n_gpus_rollout,
            replica_script="",
            control_url="",
            output_dir=None,
            script=script,
            backend=backend,
        )
        if args.num_workers is not None:
            assert args.num_workers >= len(global_launch_settings)
        num_workers = len(global_launch_settings)
        logger.info(f"Number of workers required: {num_workers}")

        # Handle workers and communication
        if num_workers > 0:
            job_spec.completions = num_workers
            job_spec.parallelism = num_workers
            job_spec.intra_job_communication = True
        elif args.lepton_intra_job_communication is not None:
            job_spec.intra_job_communication = args.lepton_intra_job_communication

        # Set failure retry settings
        if args.lepton_max_failure_retry:
            job_spec.max_failure_retry = args.lepton_max_failure_retry
        if args.lepton_max_job_failure_retry:
            job_spec.max_job_failure_retry = args.lepton_max_job_failure_retry

        # Handle command
        job_spec.container.command = ["/bin/bash", "-c", launch_cmd]

        # Set container image
        if args.lepton_container_image:
            job_spec.container.image = args.lepton_container_image
        else:
            job_spec.container.image = BASE_IMAGE

        # Handle ports
        if args.lepton_container_port:
            job_spec.container.ports = [
                make_container_port_from_string(p) for p in args.lepton_container_port
            ]

        # Handle environment variables and secrets
        if args.lepton_env or args.lepton_secret:
            job_spec.envs = make_env_vars_from_strings(
                args.lepton_env, args.lepton_secret
            )

        # Handle mounts
        if args.lepton_mount:
            job_spec.mounts = make_mounts_from_strings(args.lepton_mount)

        # Set other configurations
        if args.lepton_image_pull_secrets:
            job_spec.image_pull_secrets = args.lepton_image_pull_secrets
        if args.lepton_privileged:
            job_spec.privileged = args.lepton_privileged
        if args.lepton_ttl_seconds_after_finished:
            job_spec.ttl_seconds_after_finished = args.lepton_ttl_seconds_after_finished
        if args.lepton_log_collection is not None:
            job_spec.log = LeptonLog(enable_collection=args.lepton_log_collection)
        if args.lepton_shared_memory_size is not None:
            job_spec.shared_memory_size = args.lepton_shared_memory_size

        # Handle reservation
        if args.lepton_with_reservation:
            if not args.lepton_node_group:
                logger.error(
                    "Error: --lepton-with-reservation is only supported for dedicated node groups"
                )
                sys.exit(1)
            job_spec.reservation_config = ReservationConfig(
                reservation_id=args.lepton_with_reservation
            )

        # Create job
        job = LeptonJob(
            spec=job_spec,
            metadata=Metadata(
                id=args.lepton_job_name,
                visibility=LeptonVisibility(args.lepton_visibility)
                if args.lepton_visibility
                else None,
            ),
        )

        # Create the job
        created_job = client.job.create(job)
        new_job_id = created_job.metadata.id_
        logger.info("🎉 Job Created Successfully!")
        logger.info(f"Name: {args.lepton_job_name}")
        logger.info(f"ID: {new_job_id}")

        return

    import cosmos_rl.utils.util as util

    logger.info(
        f"Number of policy replicas: {n_policy} with {min_n_gpus_policy} gpus each"
    )
    logger.info(
        f"Number of rollout replicas: {n_rollouts} with {min_n_gpus_rollout} gpus each"
    )

    # Get available GPUs
    available_gpus = get_available_gpus()
    if not available_gpus:
        raise RuntimeError("No GPUs available. Please check your GPU configuration.")

    # List of bash scripts to run (these should exist in the same directory)
    script_names = ["launch_controller.sh", "launch_replica.sh"]

    # Get the directory where this script is located
    script_dir = os.path.dirname(os.path.abspath(__file__))

    # Verify scripts exist and are executable
    for script_name in script_names:
        script_path = os.path.join(script_dir, script_name)
        if not os.path.exists(script_path):
            logger.error(f"Error: Script {script_path} does not exist")
            sys.exit(1)
        if not os.access(script_path, os.X_OK):
            logger.error(f"Error: Script {script_path} is not executable")
            sys.exit(1)

    controller_script = os.path.join(script_dir, "launch_controller.sh")
    replica_script = os.path.join(script_dir, "launch_replica.sh")

    # Create commands for controller
    if args.log_dir is not None:
        output_dir = args.log_dir
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        timestamp = time.strftime("%Y%m%d-%H%M%S")
        output_dir = os.path.join(output_dir, f"logs_{timestamp}")
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
    else:
        output_dir = None

    def resolve_host(host):
        try:
            result = subprocess.run(
                ["getent", "hosts", "--", host],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )

            if result.returncode == 0:
                if len(result.stdout.strip().split()) > 0:
                    return result.stdout.strip().split()[0]
                else:
                    return None
            else:
                raise RuntimeError(f"Resolution failed: {result.stderr}")
        except subprocess.TimeoutExpired:
            raise TimeoutError("DNS resolution timed out")

    def resolve_host_blocking(hostname):
        try:
            while True:
                new_hostname = resolve_host(hostname)
                if new_hostname is not None:
                    hostname = new_hostname
                    logger.info(f"Resolved hostname: {hostname}")
                    break
                time.sleep(1)
        except Exception:
            pass
        return hostname

    if (
        "LEPTON_JOB_WORKER_INDEX" in os.environ
        and int(os.environ.get("LEPTON_JOB_WORKER_INDEX")) >= 0
    ):
        cur_work_idx = int(os.environ.get("LEPTON_JOB_WORKER_INDEX"))
    else:
        cur_work_idx = args.worker_idx

    if "LEPTON_JOB_WORKER_INDEX" in os.environ:
        prefix = os.environ.get(
            "LEPTON_JOB_SERVICE_PREFIX", os.environ.get("LEPTON_JOB_NAME")
        )
        subdomain = os.environ.get("LEPTON_SUBDOMAIN", "")
        hostname = f"{prefix}-{cur_work_idx}.{subdomain}"
        import cosmos_rl.utils.network_util as network_util

        ips = network_util.get_eth_ips()
        assert len(ips) > 0, "No IPs found for the current machine"
        logger.info(
            f"Setting hostname to {hostname} {ips[0]} for worker index {cur_work_idx}"
        )
        os.system(f"hostname {ips[0]}")

    control_url = None
    if args.url is not None:
        ip, port = args.url.split(":")
        if ip in get_local_ip():
            # If the IP is the local IP, launch the controller on the local machine
            port = util.find_available_port(int(port))
            logger.info(f"Using local IP: {ip} so launching controller on port {port}")
        else:
            control_url = args.url
    else:
        if (
            "LEPTON_JOB_WORKER_INDEX" in os.environ
            and int(os.environ.get("LEPTON_JOB_WORKER_INDEX")) != 0
        ):
            # For non-primary workers, connect to the primary worker (index 0) using its hostname
            prefix = os.environ.get(
                "LEPTON_JOB_SERVICE_PREFIX", os.environ.get("LEPTON_JOB_NAME")
            )
            subdomain = os.environ.get("LEPTON_SUBDOMAIN", "")
            primary_hostname = f"{prefix}-0.{subdomain}"
            primary_hostname = resolve_host_blocking(primary_hostname)
            control_url = f"{primary_hostname}:{args.port}"
        elif "LEPTON_JOB_WORKER_INDEX" in os.environ:
            # If we're in a Lepton job prime node, check if the port is available
            if not util.is_port_free(args.port):
                raise RuntimeError(f"Port {args.port} is not available")
            else:
                port = args.port
        else:
            port = util.find_available_port(args.port)

    if control_url is None:
        logger.info(f"Controller will be launched locally on port {port}.")
    else:
        logger.info(
            f"Controller will be launched on another node. This node will connect to {control_url} for control."
        )

    controller_cmd = None
    tmpfile_toml = None
    if n_policy > 0 or n_rollouts > 0:
        # Do not update the config if no replicas are needed which means launch controller only.
        if "policy" in cosmos_config and "parallelism" in cosmos_config["policy"]:
            cosmos_config["policy"]["parallelism"]["n_init_replicas"] = n_policy
        if "rollout" in cosmos_config and "parallelism" in cosmos_config["rollout"]:
            # Only available for RL.
            cosmos_config["rollout"]["parallelism"]["n_init_replicas"] = n_rollouts
    # Create a temporary file and write to it
    with tempfile.NamedTemporaryFile(
        mode="w+", suffix=".toml", delete=False
    ) as tmpfile:
        toml.dump(cosmos_config, tmpfile)
        tmpfile_toml = tmpfile.name

    if control_url is None:
        logger.info(f"Temporary configuration file created at {tmpfile_toml}")
        controller_cmd = f"{controller_script} --config {tmpfile_toml}"
        controller_cmd += f" --port {port}"
        if script:
            controller_cmd += f" {script}"
        control_url = f"localhost:{port}"

    def get_lepton_ip(worker_idx: int) -> str:
        if "LEPTON_JOB_WORKER_INDEX" in os.environ:
            # For non-primary workers, connect to the primary worker (index 0) using its hostname
            prefix = os.environ.get(
                "LEPTON_JOB_SERVICE_PREFIX", os.environ.get("LEPTON_JOB_NAME")
            )
            subdomain = os.environ.get("LEPTON_SUBDOMAIN", "")
            hostname = f"{prefix}-{worker_idx}.{subdomain}"
            hostname = resolve_host_blocking(hostname)
        else:
            raise RuntimeError(
                "Lepton job worker index not found in environment variables"
            )
        return hostname

    def get_ip_from_list(worker_idx: int) -> str:
        if args.node_ip_list is not None:
            logger.info(f"Node IP list provided: {args.node_ip_list}")
            ip_list = args.node_ip_list.split(";")
            logger.info(f"Node IP list: {ip_list}")
            if worker_idx < len(ip_list):
                return ip_list[worker_idx]
            else:
                raise RuntimeError(
                    f"Worker index {worker_idx} exceeds the length of the IP list"
                )
        else:
            raise RuntimeError("Node IP list not provided")

    def get_worker_ip(worker_idx: int) -> str:
        if "LEPTON_JOB_WORKER_INDEX" in os.environ:
            return get_lepton_ip(worker_idx)
        elif args.node_ip_list is not None:
            return get_ip_from_list(worker_idx)
        else:
            raise RuntimeError(
                "Replica with GPUs larger than 8 occurs but not on Lepton job, please specify --node-ip-list to provide the IPs of all nodes to enable conenctions to each Rendezvous head node."
            )

    global_launch_settings = replica_placement(
        available_gpus,
        n_policy,
        n_rollouts,
        min_n_gpus_policy,
        min_n_gpus_rollout,
        replica_script,
        control_url,
        output_dir,
        get_worker_ip=get_worker_ip,
        rdzv_port=args.rdzv_port,
        script=script,
        backend=backend,
        config_path=tmpfile_toml,
    )

    num_workers = len(global_launch_settings)
    logger.info(f"Number of workers required: {num_workers}")
    if num_workers > 1:
        logger.info(
            "Multiple worker nodes will be used. Ensure that the launch script is excuted on all worker nodes."
        )
    assert (
        len(available_gpus) * num_workers
        >= min_n_gpus_policy * n_policy + min_n_gpus_rollout * n_rollouts
    ), f"Not enough GPUs available. Required: {min_n_gpus_policy * n_policy + min_n_gpus_rollout * n_rollouts}, Available: {len(available_gpus)}"

    if (
        len(global_launch_settings) <= cur_work_idx
        or len(global_launch_settings[cur_work_idx]) == 0
    ):
        if controller_cmd is None:
            logger.info(
                f"No launch settings found for worker index {cur_work_idx}, no need launch"
            )
            sys.exit(0)

    processes = []

    controller_id = -1

    if controller_cmd is not None:
        controller_process = launch_processes(
            [controller_cmd],
            [""],
            [""],
            [
                os.path.join(output_dir, "controller.log")
                if output_dir is not None
                else None
            ],
        )
        controller_id = len(processes)
        processes.append(controller_process[0])

    logger.info(f"Waiting for controller to be ready at {control_url}")
    wait_for_url_ready(
        control_url, controller_process[0] if controller_cmd is not None else None
    )
    logger.info(f"Controller is ready at {control_url}")

    if (
        len(global_launch_settings) > cur_work_idx
        and len(global_launch_settings[cur_work_idx]) != 0
    ):
        commands = global_launch_settings[cur_work_idx][0]
        gpu_devices = global_launch_settings[cur_work_idx][1]
        control_urls = global_launch_settings[cur_work_idx][2]
        output_files = global_launch_settings[cur_work_idx][3]

        # Combine all commands
        logger.info(f"Commands to be executed: {commands}")
        logger.info(f"GPU devices to be used: {gpu_devices}")
        logger.info(f"Control URLs to be used: {control_urls}")
        logger.info(f"Output files: {output_files}")

        # Check if the number of GPU devices matches the number of commands
        assert (
            len(gpu_devices) == len(commands)
        ), f"Number of GPU devices ({len(gpu_devices)}) does not match number of commands ({len(commands)})"

        # Launch all processes
        processes.extend(
            launch_processes(commands, gpu_devices, control_urls, output_files)
        )

    # Wait for all processes to complete without blocking
    while len(processes) > 0:
        for i, process in enumerate(processes):
            try:
                # Check if process has finished without blocking
                if process.poll() is not None:
                    returncode = process.returncode
                    if returncode == 0:
                        logger.info(f"Process {i} completed successfully")
                    else:
                        logger.error(
                            f"Process {i} failed with return code {returncode}"
                        )
                        # Terminate all remaining processes
                        if controller_id == -1 or i == controller_id:
                            for p in processes:
                                try:
                                    p.kill()
                                except Exception as e:
                                    logger.error(f"Error kill process {p}: {e}")
                            logger.error("Terminated all processes due to failure")
                            sys.exit(1)  # Exit with error code 1 if any process failed
                    # Remove completed process from list
                    processes.remove(process)
            except Exception as e:
                logger.error(f"Error monitoring process {i}: {e}")
                # Terminate all remaining processes
                if controller_id == -1 or i == controller_id:
                    for p in processes:
                        try:
                            p.kill()
                        except Exception as e:
                            logger.error(f"Error kill process {p}: {e}")
                    logger.error("Terminated all processes due to error")
                    sys.exit(1)
        # Small sleep to prevent busy waiting
        time.sleep(0.1)

    if tmpfile_toml is not None and os.path.exists(tmpfile_toml):
        # Clean up the temporary file
        try:
            os.unlink(tmpfile_toml)
            tmpfile_toml = None
        except Exception as e:
            logger.error(f"Error deleting temporary file {tmpfile_toml}: {e}")


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import weakref
from typing import Optional
from pydantic import Field

from cosmos_rl.policy.config import Config as CosmosConfig
from cosmos_rl.utils.logging import logger

from tensorrt_llm._torch.pyexecutor.py_executor import PyExecutor
from tensorrt_llm.lora_manager import (
    LoraConfig,
    get_default_trtllm_modules_to_hf_modules,
    load_torch_lora,
)
from tensorrt_llm.llmapi.llm_utils import print_traceback_on_error

from tensorrt_llm.inputs import create_input_processor
from tensorrt_llm.llmapi.llm_args import PybindMirror
from tensorrt_llm.llmapi.tokenizer import (
    _xgrammar_tokenizer_info,
    _llguidance_tokenizer_info,
)
from tensorrt_llm.llmapi.mpi_session import external_mpi_comm_available
from tensorrt_llm.executor import PostprocWorkerConfig
from tensorrt_llm.executor.ipc import ZeroMqQueue as IpcQueue
from tensorrt_llm.bindings import executor as tllm_executor
from tensorrt_llm._torch.pyexecutor import _util as tllm_util
from tensorrt_llm._torch.pyexecutor._util import _try_infer_num_experts
from tensorrt_llm._torch.pyexecutor.resource_manager import (
    PeftCacheManager,
    ResourceManager,
    ResourceManagerType,
)
from tensorrt_llm.llmapi.llm_args import PeftCacheConfig, NGramDecodingConfig
from tensorrt_llm._torch.pyexecutor.scheduler import (
    BindCapacityScheduler,
    BindMicroBatchScheduler,
    SimpleScheduler,
)
from tensorrt_llm._torch.pyexecutor.guided_decoder import GuidedDecoder
from tensorrt_llm._torch.pyexecutor.seq_slot_manager import SeqSlotManager
from tensorrt_llm._torch.pyexecutor.kv_cache_transceiver import (
    AttentionTypeCpp,
    create_kv_cache_transceiver,
)
from tensorrt_llm._torch.pyexecutor.config_utils import (
    is_mla,
)

from cosmos_rl.rollout.trtllm_rollout.trtllm_worker import (
    CosmosTRTLLMWorker,
    CosmosWorkerCommIpcAddrs,
)
from cosmos_rl.utils.mpi_distributed import init_distributed_with_MPI


"""
Patches for trtllm.
"""

# 1. Add cosmos config to trtllm ExecutorConfig and `LLMArgs`. patch the _build_model of `LLM` classes.

import tensorrt_llm
from tensorrt_llm.llmapi import llm_args as tllm_llm_args
import tensorrt_llm.llmapi.llm as llm_trllm


def patch_trtllm_llm_args():
    class CosmosLLMArgs(tllm_llm_args.TorchLlmArgs):
        cosmos_config: Optional[CosmosConfig] = Field(
            default=None, description="Cosmos config that hacked by cosmos-rl"
        )

    # Because trtllm in v1.0.0 use many `from import` to import the class, we need to patch all the class.
    tllm_llm_args.TorchLlmArgs = CosmosLLMArgs
    tllm_llm_args.LlmArgs = CosmosLLMArgs
    tensorrt_llm.LlmArgs = CosmosLLMArgs
    llm_trllm.LlmArgs = CosmosLLMArgs
    llm_trllm.TorchLlmArgs = CosmosLLMArgs


patch_trtllm_llm_args()


def extend_create_py_executor_instance():
    def cosmos_create_py_executor_instance(
        *,
        dist,
        resources,
        mapping,
        pytorch_backend_config,
        executor_config,
        ctx_chunk_config,
        model_engine,
        start_worker,
        sampler,
        drafter,
        guided_decoder: Optional[GuidedDecoder] = None,
        lora_config: Optional[LoraConfig] = None,
        garbage_collection_gen0_threshold: Optional[int] = None,
    ) -> PyExecutor:
        kv_cache_manager = resources.get(ResourceManagerType.KV_CACHE_MANAGER, None)

        spec_config = model_engine.spec_config

        logger.info(
            f"max_seq_len={executor_config.max_seq_len}, max_num_requests={executor_config.max_batch_size}, max_num_tokens={executor_config.max_num_tokens}, max_batch_size={executor_config.max_batch_size}"
        )

        for key, value in pytorch_backend_config.extra_resource_managers.items():
            if key in resources:
                raise ValueError(f"Cannot overwrite existing resource manager {key}.")
            resources[key] = value

        peft_cache_manager = None
        if lora_config is not None:
            from tensorrt_llm.bindings import LoraModule

            if len(lora_config.lora_dir) == 1:
                # Route to appropriate loader based on checkpoint source
                load_torch_lora(lora_config)
            else:
                assert (
                    len(lora_config.lora_target_modules) >= 1
                ), "Expecting at least one lora target module"
                if not bool(lora_config.trtllm_modules_to_hf_modules):
                    lora_config.trtllm_modules_to_hf_modules = (
                        get_default_trtllm_modules_to_hf_modules()
                    )

            model_binding_config = (
                model_engine.model.model_config.get_bindings_model_config()
            )

            num_experts = _try_infer_num_experts(model_engine.model.model_config)

            num_kv_attention_heads_per_layer = (
                model_binding_config.num_kv_heads_per_layer
            )
            if max(num_kv_attention_heads_per_layer) != min(
                num_kv_attention_heads_per_layer
            ):
                logger.warning(
                    "Defining LORA with per-layer KV heads is not supported for LORA, using the max number of KV heads per layer"
                )
                num_kv_attention_heads = max(num_kv_attention_heads_per_layer)
            else:
                # all layers have the same number of KV heads
                num_kv_attention_heads = num_kv_attention_heads_per_layer[0]

            lora_modules = LoraModule.create_lora_modules(
                lora_module_names=lora_config.lora_target_modules,
                hidden_size=model_binding_config.hidden_size,
                mlp_hidden_size=model_binding_config.mlp_hidden_size,
                num_attention_heads=model_binding_config.num_heads,
                num_kv_attention_heads=num_kv_attention_heads,
                attention_head_size=model_binding_config.head_size,
                tp_size=mapping.tp_size,
                num_experts=num_experts,
            )
            model_binding_config.use_lora_plugin = True
            model_binding_config.lora_modules = lora_modules
            model_binding_config.max_lora_rank = lora_config.max_lora_rank

            max_lora_rank = lora_config.max_lora_rank
            num_lora_modules = (
                model_engine.model.model_config.pretrained_config.num_hidden_layers
                * len(lora_config.lora_target_modules + lora_config.missing_qkv_modules)
            )

            peft_cache_config_model = (
                PeftCacheConfig.from_pybind(executor_config.peft_cache_config)
                if executor_config.peft_cache_config is not None
                else PeftCacheConfig()
            )
            if lora_config.max_loras is not None:
                peft_cache_config_model.num_device_module_layer = (
                    max_lora_rank * num_lora_modules * lora_config.max_loras
                )
            if lora_config.max_cpu_loras is not None:
                peft_cache_config_model.num_host_module_layer = (
                    max_lora_rank * num_lora_modules * lora_config.max_cpu_loras
                )
            executor_config.peft_cache_config = peft_cache_config_model._to_pybind()

            from tensorrt_llm.bindings import WorldConfig

            world_config = WorldConfig(
                tensor_parallelism=mapping.tp_size,
                pipeline_parallelism=mapping.pp_size,
                context_parallelism=mapping.cp_size,
                rank=dist.mapping.rank,
                gpus_per_node=dist.mapping.gpus_per_node,
            )
            peft_cache_manager = PeftCacheManager(
                peft_cache_config=executor_config.peft_cache_config,
                model_config=model_binding_config,
                world_config=world_config,
            )
            resources[ResourceManagerType.PEFT_CACHE_MANAGER] = peft_cache_manager
            model_engine.set_lora_model_config(
                lora_config.lora_target_modules,
                lora_config.trtllm_modules_to_hf_modules,
            )

        max_num_sequences = executor_config.max_batch_size * mapping.pp_size

        resources[ResourceManagerType.SEQ_SLOT_MANAGER] = SeqSlotManager(
            max_num_sequences
        )

        resource_manager = ResourceManager(resources)

        # Make sure the kv cache manager is always invoked last as it could
        # depend on the results of other resource managers.
        if kv_cache_manager is not None:
            resource_manager.resource_managers.move_to_end(
                ResourceManagerType.KV_CACHE_MANAGER, last=True
            )

        capacity_scheduler = BindCapacityScheduler(
            max_num_sequences,
            kv_cache_manager.impl if kv_cache_manager is not None else None,
            peft_cache_manager.impl if peft_cache_manager is not None else None,
            executor_config.scheduler_config.capacity_scheduler_policy,
            two_step_lookahead=mapping.has_pp(),
        )
        mb_scheduler = BindMicroBatchScheduler(
            executor_config.max_batch_size,
            executor_config.max_num_tokens,
            ctx_chunk_config,
        )
        scheduler = SimpleScheduler(capacity_scheduler, mb_scheduler)

        config = model_engine.model.model_config.pretrained_config
        attention_type = (
            AttentionTypeCpp.MLA if is_mla(config) else AttentionTypeCpp.DEFAULT
        )
        cache_transceiver_config = executor_config.cache_transceiver_config
        kv_cache_transceiver = create_kv_cache_transceiver(
            mapping, kv_cache_manager, attention_type, cache_transceiver_config
        )
        return CosmosTRTLLMWorker(
            resource_manager,
            scheduler,
            model_engine=model_engine,
            sampler=sampler,
            drafter=drafter,
            dist=dist,
            max_num_sequences=max_num_sequences,
            disable_overlap_scheduler=pytorch_backend_config.disable_overlap_scheduler,
            max_batch_size=executor_config.max_batch_size,
            max_beam_width=executor_config.max_beam_width,
            max_draft_len=spec_config.max_draft_len if spec_config is not None else 0,
            kv_cache_transceiver=kv_cache_transceiver,
            guided_decoder=guided_decoder,
            start_worker=start_worker,
            garbage_collection_gen0_threshold=garbage_collection_gen0_threshold,
            cosmos_config=executor_config.cosmos_config,
            cosmos_ipc_queues=executor_config.cosmos_ipc_queues_addrs,
        )

    tllm_util.create_py_executor_instance = cosmos_create_py_executor_instance


extend_create_py_executor_instance()

from tensorrt_llm.llmapi.llm import _TorchLLM


def patch_trtllm_build_model():
    def cosmos_build_model(self, *args, **kwargs):
        super(_TorchLLM, self)._build_model()
        assert self._engine_dir is None

        # Tokenizer loading should be after calling model_loader(), since model_loader() may download the model from HF hub.
        # It should also be before bindings ExecutorConfig, which may depend on tokenizer info.
        self._tokenizer = self._try_load_tokenizer()

        # Multimodal special handling:
        # 1. Default load_tokenizer may fail because MM has different tokenizer configuration. Hence we initialize it inside input processor
        # 2. May need to modify model weights for MM (e.g., resize vocab embedding). We must do such operation via input processor's __init__
        self.input_processor = create_input_processor(
            self._hf_model_dir, self.tokenizer
        )
        self.tokenizer = self.input_processor.tokenizer

        max_batch_size = self.args.max_batch_size
        max_num_tokens = self.args.max_num_tokens
        max_seq_len = self.args.max_seq_len

        kwargs = {}
        if self._on_trt_backend:
            kwargs["batching_type"] = (
                self.args.batching_type or tllm_executor.BatchingType.INFLIGHT
            )

        self._executor_config = tllm_executor.ExecutorConfig(
            max_beam_width=self.args.max_beam_width,
            scheduler_config=PybindMirror.maybe_to_pybind(self.args.scheduler_config),
            max_batch_size=max_batch_size,
            max_num_tokens=max_num_tokens,
            gather_generation_logits=self.args.gather_generation_logits,
            fail_fast_on_attention_window_too_large=getattr(
                self.args, "fail_fast_on_attention_window_too_large", False
            ),
            **kwargs,
        )

        if self.args.kv_cache_config is not None:
            self._executor_config.kv_cache_config = PybindMirror.maybe_to_pybind(
                self.args.kv_cache_config
            )
        if os.getenv("FORCE_DETERMINISTIC", "0") == "1":
            # Disable KV cache reuse for deterministic mode
            self._executor_config.kv_cache_config.enable_block_reuse = False
            self._executor_config.kv_cache_config.enable_partial_reuse = False
        if self.args.peft_cache_config is not None:
            self._executor_config.peft_cache_config = PybindMirror.maybe_to_pybind(
                self.args.peft_cache_config
            )
        if self.args.decoding_config is not None:
            self._executor_config.decoding_config = self.args.decoding_config
        if self.args.guided_decoding_backend == "xgrammar":
            self._executor_config.guided_decoding_config = tllm_executor.GuidedDecodingConfig(
                backend=tllm_executor.GuidedDecodingConfig.GuidedDecodingBackend.XGRAMMAR,
                **_xgrammar_tokenizer_info(self.tokenizer),
            )
        elif self.args.guided_decoding_backend == "llguidance":
            self._executor_config.guided_decoding_config = tllm_executor.GuidedDecodingConfig(
                backend=tllm_executor.GuidedDecodingConfig.GuidedDecodingBackend.LLGUIDANCE,
                **_llguidance_tokenizer_info(self.tokenizer),
            )
        elif self.args.guided_decoding_backend is not None:
            raise ValueError(
                f"Unsupported guided decoding backend {self.args.guided_decoding_backend}"
            )

        if self._on_trt_backend:
            self._executor_config.normalize_log_probs = self.args.normalize_log_probs
        self._executor_config.enable_chunked_context = self.args.enable_chunked_prefill
        self._executor_config.max_beam_width = self.args.max_beam_width
        if self.args.cache_transceiver_config is not None:
            self._executor_config.cache_transceiver_config = (
                PybindMirror.maybe_to_pybind(self.args.cache_transceiver_config)
            )
        from tensorrt_llm._torch.pyexecutor.config import update_executor_config

        spec_config = self.args.speculative_config
        max_batch_size = self._executor_config.max_batch_size
        # Apply default heuristic to AutoDecodingConfig based on benchmark results
        # With concurrency <= 4, max_draft_len = 5, max_matching_ngram_size = 3
        # With concurrency <= 32, max_draft_len = 3, max_matching_ngram_size = 5
        # With concurrency > 32, speculative decoding is disabled.
        if spec_config is not None and spec_config.decoding_type == "AUTO":
            if not self.args.disable_overlap_scheduler:
                logger.info(
                    "Disable overlap scheduler to enable Auto speculative decoding with Ngram."
                )
                # From benchmark results, we found that NGram speculative decoding provides better performance than overlap scheduler with low concurrency <= 32.
                # Therefore, we disable overlap scheduler to enable NGram speculative decoding.
                self.args.disable_overlap_scheduler = True

            spec_config = NGramDecodingConfig(
                max_draft_len=5 if max_batch_size <= 4 else 3,
                max_matching_ngram_size=3 if max_batch_size <= 4 else 5,
                is_keep_all=True,
                is_use_oldest=True,
                is_public_pool=True,
                # Flag to indicate the NGramDecodingConfig is instantiated by auto heuristic.
                is_auto_heuristic=True,
            )

            logger.info(
                f"Apply heuristic to incomplete NGramDecodingConfig: max_draft_len={spec_config.max_draft_len}, max_matching_ngram_size={spec_config.max_matching_ngram_size}"
            )

        update_executor_config(
            self._executor_config,
            backend=self.args.backend,
            pytorch_backend_config=self.args.get_pytorch_backend_config()
            if self.args.backend in ["pytorch", "_autodeploy"]
            else None,
            mapping=self.args.parallel_config.to_mapping(),
            speculative_config=spec_config,
            hf_model_dir=self._hf_model_dir,
            max_input_len=self.args.max_input_len,
            max_seq_len=max_seq_len,
            checkpoint_format=None
            if self.args.backend == "_autodeploy"
            else self.args.checkpoint_format,
            checkpoint_loader=None
            if self.args.backend == "_autodeploy"
            else self.args.checkpoint_loader,
        )

        """
        Cosmos-RL modification start.
        - Add cosmos config to executor_config.
        """
        logger.info("[Rollout] update executor_config.cosmos_config")
        self._executor_config.cosmos_config = self.args.cosmos_config
        # set up IPCQueues that we need.
        self.cosmos_replica_name_queue = IpcQueue(
            is_server=True, name="cosmos_replica_name_queue"
        )
        self.cosmos_weight_sync_queue = IpcQueue(
            is_server=True, name="cosmos_weight_sync_queue"
        )
        # dynamically set the ipc_queues to the executor_config.
        self._executor_config.cosmos_ipc_queues_addrs = CosmosWorkerCommIpcAddrs(
            replica_name_queue_addr=self.cosmos_replica_name_queue.address,
            weight_sync_queue_addr=self.cosmos_weight_sync_queue.address,
        )
        """
        Cosmos-RL modification end.
        """

        # TODO: revisit gather_context_logits
        return_logits = self.args.gather_generation_logits

        self._executor = self._executor_cls.create(
            self._engine_dir,
            executor_config=self._executor_config,
            batched_logits_processor=self.args.batched_logits_processor,
            model_world_size=self.args.parallel_config.world_size,
            mpi_session=self.mpi_session,
            reuse_mpi_comm=external_mpi_comm_available(
                self.args.parallel_config.world_size
            ),
            return_logits=return_logits,
            postproc_worker_config=PostprocWorkerConfig(
                num_postprocess_workers=self.args.num_postprocess_workers,
                postprocess_tokenizer_dir=self.args.postprocess_tokenizer_dir,
            ),
            is_llm_executor=True,
            lora_config=self.args.lora_config,
            garbage_collection_gen0_threshold=self.args.garbage_collection_gen0_threshold,
        )

    _TorchLLM._build_model = cosmos_build_model


patch_trtllm_build_model()


# Patch worker_main
from tensorrt_llm.executor.worker import worker_main


@print_traceback_on_error
def cosmos_worker_main(*args, **kwargs) -> None:
    origin_worker_main = worker_main
    # init torch distributed environment
    init_distributed_with_MPI()

    return origin_worker_main(*args, **kwargs)


def patch_worker_main():
    import torch
    from tensorrt_llm.executor.proxy import GenerationExecutorProxy
    import concurrent.futures
    from tensorrt_llm.llmapi.tracer import enable_llm_tracer, get_tracer

    def cosmos_start_executor_workers(self, worker_kwargs):
        self_ref = weakref.ref(self)

        def mpi_done_callback(future: concurrent.futures.Future):
            # This is called when the MPI worker is done, so future.exception()
            # will not block.
            if future.exception() is not None:
                if self_ := self_ref():
                    self_._error_queue.put_nowait(future.exception())

        tracer_init_kwargs = get_tracer().init_kwargs if enable_llm_tracer() else None
        from tensorrt_llm._torch.models.modeling_auto import MODEL_CLASS_MAPPING

        torch.cuda.Stream()
        self.mpi_futures = self.mpi_session.submit(
            cosmos_worker_main,
            **worker_kwargs,
            worker_cls=self.worker_cls,
            tracer_init_kwargs=tracer_init_kwargs,
            _torch_model_class_mapping=MODEL_CLASS_MAPPING,
            ready_signal=GenerationExecutorProxy.READY_SIGNAL,
        )
        for fut in self.mpi_futures:
            fut.add_done_callback(mpi_done_callback)

        self.workers_started = True

        while True:
            if self.worker_init_status_queue.poll(1):
                ready_signal = self.worker_init_status_queue.get()
                break
            if any(fut.done() for fut in self.mpi_futures):
                logger.error("Executor worker died during initialization.")
                raise RuntimeError("Executor worker died during initialization")
            self._handle_background_error()

        if ready_signal != GenerationExecutorProxy.READY_SIGNAL:
            self.mpi_session.shutdown_abort(reason=ready_signal)
            raise RuntimeError("Executor worker returned error") from ready_signal
        self_ref = weakref.ref(self)

        def mpi_done_callback(future: concurrent.futures.Future):
            # This is called when the MPI worker is done, so future.exception()
            # will not block.
            if future.exception() is not None:
                if self_ := self_ref():
                    self_._error_queue.put_nowait(future.exception())

    GenerationExecutorProxy.worker_main = cosmos_worker_main
    GenerationExecutorProxy._start_executor_workers = cosmos_start_executor_workers


patch_worker_main()


# patch the ExecutorRequestQueue
import datetime
import queue
from typing import List
from tensorrt_llm._torch.pyexecutor.executor_request_queue import (
    ExecutorRequestQueue,
    RequestQueueItem,
)


def cosmos_get_from_request_queue(
    self, timeout: Optional[datetime.timedelta]
) -> List[RequestQueueItem]:
    items = []
    timeout_secs = timeout.total_seconds() if timeout is not None else None
    try:
        if self.request_queue.empty() and (timeout_secs is None or timeout_secs > 0):
            # if queue is empty and want to wait, wait
            # items.append(self.request_queue.get(timeout=timeout_secs))

            #### Cosmos-RL modification start.

            # Note: self.request_queue.get will block the thread.
            # We do not want it to block the thread.
            request = self.request_queue.get(False)
            items.append(request)

            #### Cosmos-RL modification end.
        else:
            # if not empty or don't want to wait, just return all items in queue
            while True:
                queue_item = self.request_queue.get_nowait()
                items.append(queue_item)
    except queue.Empty:
        pass
    return items


def patch_executor_request_queue():
    ExecutorRequestQueue._get_from_request_queue = cosmos_get_from_request_queue


patch_executor_request_queue()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import triton
import triton.language as tl


@triton.jit
def get_tid():
    return tl.inline_asm_elementwise(
        """
        mov.u32 $0, %tid.x;
        mov.u32 $1, %tid.y;
        mov.u32 $2, %tid.z;
        """,
        "=r,=r,=r",
        [],
        dtype=(tl.uint32, tl.uint32, tl.uint32),
        is_pure=True,
        pack=1,
    )


@triton.jit
def get_ntid():
    return tl.inline_asm_elementwise(
        """
        mov.u32 $0, %ntid.x;
        mov.u32 $1, %ntid.y;
        mov.u32 $2, %ntid.z;
        """,
        "=r,=r,=r",
        [],
        dtype=(tl.uint32, tl.uint32, tl.uint32),
        is_pure=True,
        pack=1,
    )


@triton.jit
def get_flat_tid():
    tid_x, tid_y, tid_z = get_tid()
    ntid_x, ntid_y, _ = get_ntid()
    return tid_z * ntid_y * ntid_x + tid_y * ntid_x + tid_x


@triton.jit
def get_flat_bid():
    return (
        tl.program_id(2) * tl.num_programs(1) * tl.num_programs(0)
        + tl.program_id(1) * tl.num_programs(0)
        + tl.program_id(0)
    )


@triton.jit
def sync_threads():
    tl.inline_asm_elementwise(
        "bar.sync 0;", "=r", [], dtype=tl.int32, is_pure=False, pack=1
    )
